;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; xanana.el - JaKarTa mode for Emacs and XEmacs (20 and more).

;;    Copyright  2003-2004 Daniela Andrade and Simo Melo de Sousa all rights reserved.
;;    Copying is covered by the GNU General Public License.

;;    This program is strongly based upon Tuareg, an Emacs mode For Ocaml
;;    Please look at the Tuareg recommendation (see below).

;;    This program is free software; you can redistribute it and/or modify
;;    it under the terms of the GNU General Public License as published by
;;    the Free Software Foundation; either version 2 of the License, or
;;    (at your option) any later version.

;;    This program is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;;    GNU General Public License for more details.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;        Copyright  1997-2000 Albert Cohen, all rights reserved.
;;         Copying is covered by the GNU General Public License.

;;    This program is free software; you can redistribute it and/or modify
;;    it under the terms of the GNU General Public License as published by
;;    the Free Software Foundation; either version 2 of the License, or
;;    (at your option) any later version.

;;    This program is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;;    GNU General Public License for more details.

(defconst jakarta-mode-version "Xanana/JaKarTa Version 1.00"
  "    Copyright  2003-2004 Daniela Andrade and Simo Melo de Sousa all rights reserved.
    Copying is covered by the GNU General Public License. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                        Emacs versions support

(defconst jakarta-with-xemacs (string-match "XEmacs" emacs-version))

(defconst jakarta-window-system
  (or (and (boundp 'window-system) window-system)
      (and (fboundp 'console-type) (or (eq (console-type) 'x)
				       (eq (console-type) 'gtk)
				       (eq (console-type) 'win32))))
  "Are we running under a window system?")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                      Compatibility functions

(if (fboundp 'match-string-no-properties)
    (defalias 'jakarta-match-string 'match-string-no-properties)
  (defalias 'jakarta-match-string 'match-string))

(require 'cl)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                       User customizable variables

;; use the standard `customize' interface or `jakarta-mode-hook' to
;; configure these variables

(require 'custom)

(defgroup jakarta nil
  "Support for the Jakarta language."
  :group 'languages)

;; comments

(defcustom jakarta-indent-leading-comments t
  "*If true, indent leading comment lines (starting with `(*') like others."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-indent-comments t
  "*If true, automatically align multi-line comments."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-comment-end-extra-indent 0
  "*How many spaces to indent a leading comment end `*)'.
If you expect comments to be indented like
	(*
          ...
	 *)
even without leading `*', use `jakarta-comment-end-extra-indent' = 1."
  :group 'jakarta
  :type '(radio :extra-offset 8
		:format "%{Comment End Extra Indent%}:
   Comment alignment:\n%v"
		(const :tag "align with `(' in comment opening" 0)
		(const :tag "align with `*' in comment opening" 1)
		(integer :tag "custom alignment" 0)))

(defcustom jakarta-support-leading-star-comments t
  "*Enables automatic intentation of comments of the form
        (*
         * ...
         *)
If you still expect comments to be indented like
	(*
          ...
	 *)
without leading `*', set `jakarta-comment-end-extra-indent' to 1."
  :group 'jakarta :type 'boolean)

;; indentation defaults

(defcustom jakarta-default-indent 2
  "*Default indentation.

Global indentation variable (large values may lead to indentation overflows).
When no governing keyword is found, this value is used to indent the line
if it has to."
  :group 'jakarta :type 'integer)

(defcustom jakarta-lazy-= nil
  "*If true, indent `=' like a standard keyword (not `:=', `<='...)."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-lazy-paren nil
  "*If true, indent parentheses like a standard keyword."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-support-jakartalight nil
  "*If true, handle Jakarta Light character syntax (incompatible with labels)."
  :group 'jakarta :type 'boolean
  :set '(lambda (var val)
	  (setq jakarta-support-jakartalight val)
	  (if (boundp 'jakarta-mode-syntax-table)
	      (if val
		  (modify-syntax-entry ?` "\"" jakarta-mode-syntax-table)
		(modify-syntax-entry ?` "." jakarta-mode-syntax-table)))))
(make-variable-buffer-local 'jakarta-support-jakartalight)

(defcustom jakarta-let-always-indent t
  "*If true, enforce indentation is at least `jakarta-let-indent' after a `let'.

As an example, set it to false when you have `jakarta-with-indent' set to 0,
and you want `let x = match ... with' and `match ... with' indent the
same way."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-|-extra-unindent jakarta-default-indent
  "*Extra backward indent for Jakarta lines starting with the `|' operator.

It is automatically added to `function', `with', `parse' and some cases
of `type' keywords to leave enough space for `|' backward indentation."
  :group 'jakarta :type 'integer)

(defcustom jakarta-class-indent jakarta-default-indent
  "*How many spaces to indent from a `class' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-sig-struct-align t
  "*Align `sig' and `struct' keywords with `module'."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-sig-struct-indent jakarta-default-indent
  "*How many spaces to indent from a `sig' or `struct' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-method-indent jakarta-default-indent
  "*How many spaces to indent from a `method' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-begin-indent jakarta-default-indent
  "*How many spaces to indent from a `begin' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-for-while-indent jakarta-default-indent
  "*How many spaces to indent from a `for' or `while' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-do-indent jakarta-default-indent
  "*How many spaces to indent from a `do' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-fun-indent jakarta-default-indent
  "*How many spaces to indent from a `fun' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-function-indent jakarta-default-indent
  "*How many spaces to indent from a `function' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-if-then-else-indent jakarta-default-indent
  "*How many spaces to indent from an `if', `then' or `else' keyword
in Jakarta mode."
  :group 'jakarta :type 'integer)

(defcustom jakarta-let-indent jakarta-default-indent
  "*How many spaces to indent from a `let' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-in-indent jakarta-default-indent
  "*How many spaces to indent from a `in' keyword.
A lot of people like formatting `let' ... `in' expressions whithout
indentation:
        let x = 0 in
        blah x
Set this variable to 0 to get this behaviour.
However, nested declarations are always correctly handled:
        let x = 0 in                             let x = 0
        let y = 0 in              or             in let y = 0
        let z = 0 ...                            in let z = 0 ..."
  :group 'jakarta :type 'integer)

(defcustom jakarta-match-indent jakarta-default-indent
  "*How many spaces to indent from a `match' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-try-indent jakarta-default-indent
  "*How many spaces to indent from a `try' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-with-indent jakarta-default-indent
  "*How many spaces to indent from a `with' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-rule-indent jakarta-default-indent
  "*How many spaces to indent from a `rule' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-parse-indent jakarta-default-indent
  "*How many spaces to indent from a `parse' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-parser-indent jakarta-default-indent
  "*How many spaces to indent from a `parser' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-type-indent jakarta-default-indent
  "*How many spaces to indent from a `type' keyword."
  :group 'jakarta :type 'integer)

(defcustom jakarta-val-indent jakarta-default-indent
  "*How many spaces to indent from a `val' keyword."
  :group 'jakarta :type 'integer)

;; automatic indentation
;; using abbrev-mode and electric keys

(defcustom jakarta-use-abbrev-mode t
  "*Non-nil means electrically indent lines starting with leading keyword
such as `end', `done', `else' etc. It makes use of abbrev-mode.

Many people find eletric keywords irritating, so you can disable them in
setting this variable to nil."
  :group 'jakarta :type 'boolean
  :set '(lambda (var val)
	  (setq jakarta-use-abbrev-mode val)
	  (abbrev-mode val)))
(make-variable-buffer-local 'jakarta-use-abbrev-mode)

(defcustom jakarta-electric-indent t
  "*Non-nil means electrically indent lines starting with `|', `)', `]' or `}'.

Many people find eletric keys irritating, so you can disable them in
setting this variable to nil."
  :group 'jakarta :type 'boolean)
(make-variable-buffer-local 'jakarta-electric-indent)

(defcustom jakarta-electric-close-vector t
  "*Non-nil means electrically insert `|' before a vector-closing `]' or
`>' before an object-closing `}'.

Many people find eletric keys irritating, so you can disable them in
setting this variable to nil. You should probably have this on,
though, if you also have jakarta-electric-indent on."
  :group 'jakarta :type 'boolean)
(make-variable-buffer-local 'jakarta-electric-close-vector)

;; Jakarta-Interactive
;; configure via `jakarta-mode-hook'

(defcustom jakarta-skip-after-eval-phrase t
  "*Non-nil means skip to the end of the phrase after evaluation in the
Jakarta toplevel."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-interactive-read-only-input nil
  "*Non-nil means input send to the Jakarta toplevel is read-only."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-interactive-echo-phrase t
  "*Non-nil means echo phrases in the toplevel buffer when sending
them to the Jakarta toplevel."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-interactive-input-font-lock t
  "*Non nil means Font-Lock for toplevel input phrases."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-interactive-output-font-lock t
  "*Non nil means Font-Lock for toplevel output messages."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-interactive-error-font-lock t
  "*Non nil means Font-Lock for toplevel error messages."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-display-buffer-on-eval t
  "*Non nil means pop up the Jakarta-toplevel when evaluating code."
  :group 'jakarta :type 'boolean)

(defcustom jakarta-manual-url "http://www.di.ubi.pt/~desousa/jakarta"
  "*URL to the Jakarta reference manual."
  :group 'jakarta :type 'string)

(defcustom jakarta-browser 'jakarta-netscape-manual
  "*Name of function that displays the Jakarta reference manual.
Valid names are `jakarta-netscape-manual', `jakarta-mmm-manual'
and `jakarta-xemacs-w3-manual' (XEmacs only)."
  :group 'jakarta)

(defcustom jakarta-library-path "/usr/local/lib/jakarta/"
  "*Path to the Jakarta library."
  :group 'jakarta :type 'string)

(defcustom jakarta-definitions-max-items 30
  "*Maximum number of items a definitions menu can contain."
  :group 'jakarta :type 'integer)

(defvar jakarta-options-list
  '(("Lazy parentheses indentation" . 'jakarta-lazy-paren)
    ("Lazy `=' indentation" . 'jakarta-lazy-=)
    ("Force indentation after `let'" . 'jakarta-let-always-indent)
    "---"
    ("Automatic indentation of leading keywords" . 'jakarta-use-abbrev-mode)
    ("Electric indentation of ), ] and }" . 'jakarta-electric-indent)
    ("Electric matching of [| and {<" . 'jakarta-electric-close-vector)
    "---"
    ("Indent body of comments" . 'jakarta-indent-comments)
    ("Indent first line of comments" . 'jakarta-indent-leading-comments)
    ("Leading-`*' comment style" . 'jakarta-support-leading-star-comments))
  "*List of menu-configurable Jakarta options")

(defvar jakarta-interactive-options-list
  '(("Skip phrase after evaluation" . 'jakarta-skip-after-eval-phrase)
    ("Echo phrase in interactive buffer" . 'jakarta-interactive-echo-phrase)
    "---"
    ("Font-lock interactive input" . 'jakarta-interactive-input-font-lock)
    ("Font-lock interactive output" . 'jakarta-interactive-output-font-lock)
    ("Font-lock interactive error" . 'jakarta-interactive-error-font-lock)
    "---"
    ("Read only input (XEmacs)" . 'jakarta-interactive-read-only-input))
  "*List of menu-configurable Jakarta options")

(defvar jakarta-interactive-program "jakarta -i"
  "*Default program name for invoking a Jakarta toplevel from Emacs.")
(make-variable-buffer-local 'jakarta-interactive-program)

(defvar jakarta-compile-program "jakarta "
  "*Default program name for invoking a Jakarta compiler from Emacs.")
(make-variable-buffer-local 'jakarta-compile-program)


(defvar jakarta-jakartadoc-program "jakartadoc "
  "*Default program name for invoking a Jakarta compiler from Emacs.")
(make-variable-buffer-local 'jakarta-jakartadoc-program)
;; customizable faces for Font-Lock mode

(defgroup jakarta-faces nil
  "Special faces for the Jakarta mode.

Face description is the following:
  color for LIGHT backgrounds,
  color for DARK backgrounds,
  ITALIC font, font+color fontification mode,
  BOLD font, font+color fontification mode,
  ITALIC font, font-only fontification mode,
  BOLD font, font-only fontification mode."
  :group 'jakarta)

(defcustom jakarta-font-lock-governing '("darkorange3" "orange" nil t nil t)
  "Face description for governing/leading keywords."
  :group 'jakarta-faces)

(defcustom jakarta-font-lock-operator '("brown" "khaki" nil nil nil nil)
  "Face description for all toplevel errors."
  :group 'jakarta-faces)

(defcustom jakarta-font-lock-interactive-output '("blue4" "cyan" nil nil t nil)
  "Face description for all toplevel outputs."
  :group 'jakarta-faces)

(defcustom jakarta-font-lock-interactive-error '("firebrick" "plum1" t t t t)
  "Face description for all toplevel errors."
  :group 'jakarta-faces)


(make-face 'jakartas-rule-name-face)

(make-face 'jakartas-logical-command-face)

(custom-set-faces
 '(jakartas-rule-name-face ((((type x) (class color) (background light))    
(:foreground "forestgreen" :italic t))))

'(jakartas-logical-command-face ((((type x) (class color) (background
light))     (:foreground "darkslateblue" :italic t :bold t))))
 
)


;; end of customizable variables

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            Support definitions

(defvar jakarta-cache-stop (point-min))
(make-variable-buffer-local 'jakarta-cache-stop)
(defvar jakarta-cache nil)
(make-variable-buffer-local 'jakarta-cache)
(defvar jakarta-cache-local nil)
(make-variable-buffer-local 'jakarta-cache-local)
(defvar jakarta-cache-last-local nil)
(make-variable-buffer-local 'jakarta-cache-last-local)
(defvar jakarta-last-loc (cons nil nil))

(defun jakarta-leading-star-p ()
  (and jakarta-support-leading-star-comments
       (save-excursion ; this function does not make sense outside of a comment
	 (jakarta-beginning-of-literal-or-comment)
	 (and (not (looking-at "(\\*[Tt][Ee][Xx]\\|(\\*\\*"))
	      (progn
		(forward-line 1)
		(back-to-indentation)
		(looking-at "\\*[^)]"))))))

(defun jakarta-auto-fill-insert-leading-star (&optional leading-star)
  (let ((point-leading-comment (looking-at "(\\*")) (return-leading nil))
    (save-excursion
      (back-to-indentation)
      (if jakarta-electric-indent
	  (prog2
	      (if (and (jakarta-in-comment-p)
		       (or leading-star
			   (jakarta-leading-star-p)))
		  (prog2
		      (if (not (looking-at "(?\\*"))
			  (insert-before-markers "* "))
		      (setq return-leading t)))
	      (if (not point-leading-comment)
		  ;; use optional argument to break recursion
		  (jakarta-indent-command t)))))
    return-leading))

(defun jakarta-auto-fill-function ()
  (if (jakarta-in-literal-p) ()
    (let ((leading-star
	   (if (not (char-equal ?\n last-command-char))
	       (jakarta-auto-fill-insert-leading-star)
	     nil)))
      (do-auto-fill)
      (if (not (char-equal ?\n last-command-char))
	  (jakarta-auto-fill-insert-leading-star leading-star)))))

(defun jakarta-forward-char (&optional step)
  (if step (goto-char (+ (point) step))
    (goto-char (1+ (point)))))

(defun jakarta-backward-char (&optional step)
  (if step (goto-char (- (point) step))
    (goto-char (1- (point)))))

(defun jakarta-in-indentation-p ()
  "Tests whether all characters between beginning of line and point
are blanks."
  (save-excursion
    (skip-chars-backward " \t")
    (bolp)))

(defun jakarta-before-change-function (begin end)
  (setq jakarta-cache-stop (min jakarta-cache-stop (1- begin))))

(defun jakarta-in-literal-p ()
  "Returns non-nil if point is inside a Jakarta literal."
  (car (jakarta-in-literal-or-comment)))
(defun jakarta-in-comment-p ()
  "Returns non-nil if point is inside a Jakarta comment."
  (cdr (jakarta-in-literal-or-comment)))
(defun jakarta-in-literal-or-comment-p ()
  "Returns non-nil if point is inside a Jakarta literal or comment."
  (jakarta-in-literal-or-comment)
  (or (car jakarta-last-loc) (cdr jakarta-last-loc)))
(defun jakarta-in-literal-or-comment ()
  "Returns the pair `((jakarta-in-literal-p) . (jakarta-in-comment-p))'."
  (if (and (<= (point) jakarta-cache-stop) jakarta-cache)
      (progn
	(if (or (not jakarta-cache-local) (not jakarta-cache-last-local)
		(and (>= (point) (caar jakarta-cache-last-local))))
	    (setq jakarta-cache-local jakarta-cache))
	(while (and jakarta-cache-local (< (point) (caar jakarta-cache-local)))
	  (setq jakarta-cache-last-local jakarta-cache-local
		jakarta-cache-local (cdr jakarta-cache-local)))
	(setq jakarta-last-loc
	      (if jakarta-cache-local
		  (cons (eq (cadar jakarta-cache-local) 'b)
			(> (cddar jakarta-cache-local) 0))
		(cons nil nil))))
    (let ((flag t) (op (point)) (mp (min (point) (1- (point-max))))
	  (balance 0) (end-of-comment nil))
      (while (and jakarta-cache (<= jakarta-cache-stop (caar jakarta-cache)))
	(setq jakarta-cache (cdr jakarta-cache)))
      (if jakarta-cache
	  (if (eq (cadar jakarta-cache) 'b)
	      (progn
		(setq jakarta-cache-stop (1- (caar jakarta-cache)))
		(goto-char jakarta-cache-stop)
		(setq balance (cddar jakarta-cache))
		(setq jakarta-cache (cdr jakarta-cache)))
	    (setq balance (cddar jakarta-cache))
	    (setq jakarta-cache-stop (caar jakarta-cache))
	    (goto-char jakarta-cache-stop)
	    (skip-chars-forward "("))
	(goto-char jakarta-cache-stop))
      (skip-chars-backward "\\\\*")
      (while flag
	(if end-of-comment (setq balance 0 end-of-comment nil))
	(skip-chars-forward "^\\\\'`\"(\\*")
	(cond
	 ((looking-at "\\\\")
	  (jakarta-forward-char 2))
	 ((looking-at "'\\([^\n']\\|\\\\..?.?\\)'")
	  (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'b balance))
				   jakarta-cache))
	  (skip-chars-forward "^'") (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'e balance))
				   jakarta-cache)))
	 ((and jakarta-support-jakartalight
	       (looking-at "`\\([^\n']\\|\\\\..?.?\\)`"))
	  (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'b balance))
				   jakarta-cache))
	  (skip-chars-forward "^`") (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'e balance))
				   jakarta-cache)))
	 ((looking-at "\"")
	  (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'b balance))
				   jakarta-cache))
	  (skip-chars-forward "^\\\\\"")
	  (while (looking-at "\\\\")
	    (jakarta-forward-char 2) (skip-chars-forward "^\\\\\""))
	  (jakarta-forward-char)
	  (setq jakarta-cache (cons (cons (point) (cons 'e balance))
				   jakarta-cache)))
	 ((looking-at "(\\*")
	  (setq balance (1+ balance))
	  (setq jakarta-cache (cons (cons (point) (cons nil balance))
				   jakarta-cache))
	  (jakarta-forward-char 2))
	 ((looking-at "\\*)")
	  (jakarta-forward-char 2)
	  (if (> balance 1)
	      (prog2
		  (setq balance (1- balance))
		  (setq jakarta-cache (cons (cons (point) (cons nil balance))
					   jakarta-cache)))
	    (setq end-of-comment t)
	    (setq jakarta-cache (cons (cons (point) (cons nil 0))
				     jakarta-cache))))
	 (t (jakarta-forward-char)))
	(setq flag (<= (point) mp)))
      (setq jakarta-cache-local jakarta-cache
	    jakarta-cache-stop (point))
      (goto-char op)
      (if jakarta-cache (jakarta-in-literal-or-comment) 
	(setq jakarta-last-loc (cons nil nil))
	jakarta-last-loc))))

(defun jakarta-beginning-of-literal-or-comment ()
  "Skips to the beginning of the current literal or comment (or buffer)."
  (interactive)
  (if (jakarta-in-literal-or-comment-p)
      (jakarta-beginning-of-literal-or-comment-fast)))

(defun jakarta-beginning-of-literal-or-comment-fast ()
  (while (and jakarta-cache-local
	      (or (eq 'b (cadar jakarta-cache-local))
		  (> (cddar jakarta-cache-local) 0)))
    (setq jakarta-cache-last-local jakarta-cache-local
	  jakarta-cache-local (cdr jakarta-cache-local)))
  (if jakarta-cache-last-local
      (goto-char (caar jakarta-cache-last-local))
    (goto-char (point-min)))
  (if (eq 'b (cadar jakarta-cache-last-local)) (jakarta-backward-char)))

(defun jakarta-false-=-p ()
  "Is the underlying `=' the first/second letter of an operator?"
  (or (char-equal ?: (preceding-char))
      (char-equal ?> (preceding-char))
      (char-equal ?< (preceding-char))
      (char-equal ?= (preceding-char))
      (char-equal ?= (char-after (1+ (point))))))

(defun jakarta-at-phrase-break-p ()
  "Is the underlying `;' a phrase break?"
  (and (char-equal ?\. (following-char))
       (or (and (not (eobp))
		(char-equal ?\. (char-after (1+ (point)))))
	   (char-equal ?\. (preceding-char)))))

(defun jakarta-backward-up-list ()
  "Safe up-list regarding comments, literals and errors."
  (let ((balance 1) (op (point)) (oc nil))
    (jakarta-in-literal-or-comment)
    (while (and (> (point) (point-min)) (> balance 0))
      (setq oc (if jakarta-cache-local (caar jakarta-cache-local) (point-min)))
      (condition-case nil (up-list -1) (error (goto-char (point-min))))
      (if (>= (point) oc) (setq balance (1- balance))
	(goto-char op)
	(skip-chars-backward "^[]{}()<>") (jakarta-backward-char)
	(if (not (jakarta-in-literal-or-comment-p))
	    (cond
	     ((looking-at "[[{(<]")
	      (setq balance (1- balance)))
	     ((looking-at "[]})>]")
	      (setq balance (1+ balance))))
	  (jakarta-beginning-of-literal-or-comment-fast)))
      (setq op (point)))))

(defun jakarta-assoc-indent (kwop &optional look-for-let-or-and)
  "Returns relative indentation of the keyword given in argument."
  (let ((ind (symbol-value (cdr (assoc kwop jakarta-keyword-alist))))
	(looking-let-or-and (and look-for-let-or-and
				 (looking-at "\\<\\(let\\|and\\)\\>"))))
    (if (string-match "\\<\\(with\\|fun\\(ction\\)?\\|parser?\\)\\>" kwop)
	(+ (if (and jakarta-let-always-indent
		    looking-let-or-and (< ind jakarta-let-indent))
	       jakarta-let-indent ind) jakarta-|-extra-unindent) ind)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                  Font-Lock

(if (and (featurep 'font-lock)
	 (or jakarta-window-system jakarta-with-xemacs))
    (progn
      (defun jakarta-fontify-buffer ()
	(font-lock-default-fontify-buffer)
	(jakarta-fontify (point-min) (point-max)))

      (defun jakarta-fontify-region (begin end &optional verbose)
	(font-lock-default-fontify-region begin end verbose)
	(jakarta-fontify begin end))

      (defun jakarta-after-fontify-buffer () ; compatibility with XEmacs 20.x
	(jakarta-fontify (point-min) (point-max)))

      (defun jakarta-fontify (begin end)
	(if (eq major-mode 'jakarta-mode)
	    (save-excursion
	      (let ((modified (buffer-modified-p))) ; Emacs hack (see below)
		(goto-char begin)
		(beginning-of-line)
		(setq begin (point))
		(goto-char (1- end))
		(end-of-line)

		;; dirty hack: fontify multi-line comments and
		;; trick `font-lock-default-unfontify-region'
		(forward-line 2)

		(setq end (point))
		(while (> end begin)
		  (goto-char (1- end))
		  (jakarta-in-literal-or-comment)
		  (cond
		   ((cdr jakarta-last-loc)
		    (jakarta-beginning-of-literal-or-comment)
		    (put-text-property (max begin (point)) end 'face
				       (if (looking-at
					    "(\\*[Tt][Ee][Xx]\\|(\\*\\*[^*]")
					   jakarta-doc-face
					 'font-lock-comment-face))
		    (setq end (1- (point))))
		   ((car jakarta-last-loc)
		    (jakarta-beginning-of-literal-or-comment)
		    (put-text-property (max begin (point)) end 'face
				       'font-lock-string-face)
		    (setq end (point)))
		   (t (while (and jakarta-cache-local
				  (or (> (caar jakarta-cache-local) end)
				      (eq 'b (cadar jakarta-cache-local))))
			(setq jakarta-cache-local (cdr jakarta-cache-local)))
		      (setq end (if jakarta-cache-local
				    (caar jakarta-cache-local) begin)))))
		(if (not (or jakarta-with-xemacs modified)) ; properties taken
		    (set-buffer-modified-p nil))))))       ; too seriously...

      (defun jakarta-font-lock-hook ()
	"Function called by `font-lock-mode' for initialization purposes."
	(if (eq major-mode 'jakarta-mode)
	    (jakarta-set-font-lock-faces)))

      (defun jakarta-make-face-italic (face)
	(condition-case nil (make-face-italic face) (error nil)))
      (defun jakarta-make-face-bold (face)
	(condition-case nil (make-face-bold face) (error nil)))
      (defun jakarta-make-face-unitalic (face)
	(condition-case nil (make-face-unitalic face) (error nil)))
      (defun jakarta-make-face-unbold (face)
	(condition-case nil (make-face-unbold face) (error nil)))

      (defun jakarta-set-font-lock-faces ()
	"Set faces for Font-Lock mode."
	(let* ((use-fonts
		(or (and (boundp 'font-lock-use-fonts)
			 font-lock-use-fonts
			 (not (and (boundp 'font-lock-use-colors)
				   font-lock-use-colors)))
		    (and (fboundp 'device-class)
			 (eq (device-class) 'mono))
		    (and (not (fboundp 'device-class))
			 (fboundp 'x-display-color-p)
			 (not (x-display-color-p)))))
	       (light-bg
		(if jakarta-window-system
		    (if jakarta-with-xemacs
			(if (and (fboundp 'color-rgb-components)
				 (< (apply '+ (color-rgb-components
					       (make-color-specifier
						[default background])))
				    (* (apply '+
					      (color-rgb-components
					       (make-color-specifier "white")))
				       0.6))) nil t)
		      (let ((param (cdr (assq 'background-color
					      (frame-parameters)))))
			(cond
			 ((boundp 'font-lock-background-mode)
			  (if (eq font-lock-background-mode 'dark) nil t))
			 ((eq system-type 'ms-dos)
			  (if (string-match "light" param) t nil))
			 ((and param (fboundp 'x-color-values)
			       (< (apply '+ (x-color-values param))
				  (* (apply '+ (x-color-values "white"))
				     0.6))) nil)
			 (t t))))))
	       (default-color (if light-bg "black" "white")))
	  (mapcar (lambda (name)
		    (let ((desc (eval (intern name)))
			  (face (intern (concat name "-face"))))
		      (if (or (and (functionp 'find-face) (find-face face))
			      (facep face)) ()
			(make-face face)
			(set-face-foreground
			 face (if use-fonts default-color
				(if light-bg (nth 0 desc) (nth 1 desc))))
			(if use-fonts
			    (progn
			      (if (nth 4 desc) (jakarta-make-face-italic face)
				(jakarta-make-face-unitalic face))
			      (if (nth 5 desc) (jakarta-make-face-bold face)
				(jakarta-make-face-unbold face)))
			  (if (nth 2 desc) (jakarta-make-face-italic face)
			    (jakarta-make-face-unitalic face))
			  (if (nth 3 desc) (jakarta-make-face-bold face)
			    (jakarta-make-face-unbold face))))))
		  '("jakarta-font-lock-governing"
		    "jakarta-font-lock-operator"
		    "jakarta-font-lock-interactive-output"
		    "jakarta-font-lock-interactive-error"))))

      ; XEmacs and Emacs have different documentation faces...
      (defvar jakarta-doc-face (if (facep 'font-lock-doc-face)
				  'font-lock-doc-face
				'font-lock-doc-string-face))

      (if (facep 'font-lock-constant-face) ()
	(defvar font-lock-constant-face font-lock-reference-face)
	(copy-face font-lock-reference-face 'font-lock-constant-face))

      (defvar jakarta-font-lock-keywords
	'(
	  ("\\<\\(d\\(ata\\|efinition\\)\\|and\\|variable\\|record\\|function\\|type\\)\\>"  
	   0 'jakarta-font-lock-governing-face nil)
	  ("\\<\\(r\\(eplace\\|emove\\|eject\\|educe\\)\\|select\\|totalize\\|d\\(elete\\|rop\\)\\|coercion\\|prefix\\|lemmas\\|redundant\\|dundant\\|into\\|log\\|determine\\|in\\|to\\|by\\|replace_type\\|at\\)\\>"

	   0 'font-lock-keyword-face nil)

         ("\\<\\(function\\|variable\\|definition\\)[
\t\n]+\\([a-zA-Z]\\w*\\)\\>"
          1 'font-lock-function-name-face nil)

         ("\\({\\|;\\)[ \t\n]*\\<\\([a-zA-Z]\\w*\\)\\>[ \t\n]+:"
          2 'font-lock-function-name-face nil)
          
;;;;cacrescentei do jakarta.el  
	  ("<\\([^>-]\\|->\\|-[^>]\\)+>"
	   0 'jakartas-rule-name-face nil)

	  ("\\<\\(\\(data\\([ \t\n]+type\\)?\\)\\([ \t\n]+definition\\)?\\|record\\|definition\\|function\\|variable\\([ \t\n]+type\\)?\\|type\\)\\>[ \t]*\\(\\(\\<['_A-Za-z\277-\377~?]\\w*\\>:?[ \t]*\\)*\\)\\>"
	  6 'font-lock-type-face nil)

              ("\\<\\(transparent\\|to\\|opaque\\|addprecond\\)\\>"
              0 'jakartas-logical-command-face nil)

         ("\"[^\"]*\""  0 'font-lock-string-face nil)



	  ("\\([?~]?[_A-Za-z\277-\377]\\w*[ \t\n]*:\\)[^:>=]"
	   1 'font-lock-variable-name-face nil)
	  ("\\<exception\\>[ \t]*\\(\\<[_A-Za-z\277-\377]\\w*\\>\\)"
	   1 'font-lock-variable-name-face nil)
	  ("\\<\\(build\\|from\\|abstraction\\|abstract\\|with\\|functions\\)\\>"

	   0 'font-lock-reference-face nil)
	  ("[][;,.()|{}@^!:*=&/+~?]\\.?\\|\\.\\."
	   0 'jakarta-font-lock-operator-face nil))
	"Font-Lock patterns for Jakarta mode.")

      (if (featurep 'sym-lock)
	  ;; to change this table, xfd -fn '-adobe-symbol-*--12-*' may be
	  ;; used to determine the symbol character codes.
	  (defvar jakarta-sym-lock-keywords
	    '(("->" 0 1 174)
	      ("=>" 0 1 222)
	      ("==" 0 1 186)
	      ("[^*]\\(\\*\\)\\." 1 8 180)
	      ("\\(/\\)\\." 1 3 184)
	      ;;(":>" 0 1 202)
	      ;;(";;" 0 1 191)
	      ("\\<_\\>" 0 3 188)
	      ;;("\\<sqrt\\>" 0 3 214)
	      ;;("\\<unit\\>" 0 3 198)
	      ;; ("\\<fun\\>" 0 3 108)
	      ;;("\\<or\\>" 0 3 218)
	      ;;("\\<not\\>" 0 3 216)
              ;;("<-" 0 1 172)
	      ;; (":=" 0 1 220)
	      ;;(">=" 0 1 179)
	      ;;("<>" 0 1 185)
	      ;;("||" 0 1 218)
	      ;;("&&" 0 1 217)
	      )
             "If non nil: Overrides default Sym-Lock patterns for Jakarta."))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                    Keymap

(defvar jakarta-mode-map nil
  "Keymap used in Jakarta mode.")
(setq jakarta-mode-map (make-sparse-keymap))
(define-key jakarta-mode-map "|" 'jakarta-electric)
(define-key jakarta-mode-map ")" 'jakarta-electric-rp)
(define-key jakarta-mode-map "}" 'jakarta-electric-rc)
(define-key jakarta-mode-map "]" 'jakarta-electric-rb)
(define-key jakarta-mode-map "\t" 'jakarta-indent-command)
(define-key jakarta-mode-map "\M-q" 'jakarta-indent-phrase)
(define-key jakarta-mode-map "\C-c\C-q" 'jakarta-indent-phrase)
(define-key jakarta-mode-map "\C-c\C-a" 'jakarta-find-alternate-file)
(define-key jakarta-mode-map "\C-c\C-c" 'compile)
(define-key jakarta-mode-map "\M-\C-x" 'jakarta-eval-phrase)
(define-key jakarta-mode-map "\C-x\C-e" 'jakarta-eval-phrase)
(define-key jakarta-mode-map "\C-c\C-e" 'jakarta-eval-phrase)
(define-key jakarta-mode-map "\C-c\C-r" 'jakarta-eval-region)
(define-key jakarta-mode-map "\C-c\C-b" 'jakarta-eval-buffer)
(define-key jakarta-mode-map "\C-c\C-o" 'jakarta-compile-jakarta)
(define-key jakarta-mode-map "\C-c\C-s" 'jakarta-run-jakarta)
(define-key jakarta-mode-map "\C-c\C-j" 'jakarta-jakartadoc-jakarta)
(define-key jakarta-mode-map "\C-c\C-i" 'jakarta-interrupt-jakarta)
(define-key jakarta-mode-map "\C-c\C-k" 'jakarta-kill-jakarta)
(define-key jakarta-mode-map "\C-c\C-n" 'jakarta-next-phrase)
(define-key jakarta-mode-map "\C-c\C-p" 'jakarta-previous-phrase)
(define-key jakarta-mode-map [(control c) (control down)] 'jakarta-next-phrase)
(define-key jakarta-mode-map [(control c) (control up)] 'jakarta-previous-phrase)
(define-key jakarta-mode-map [(meta control down)]  'jakarta-next-phrase)
(define-key jakarta-mode-map [(meta control up)] 'jakarta-previous-phrase)
(define-key jakarta-mode-map [(meta control h)] 'jakarta-mark-phrase)
(define-key jakarta-mode-map "\C-c`" 'jakarta-interactive-next-error-source)
(define-key jakarta-mode-map "\C-cb" 'jakarta-insert-build-from-form)
(define-key jakarta-mode-map "\C-cd" 'jakarta-insert-data-form)
(define-key jakarta-mode-map "\C-ch" 'jakarta-insert-type-form)
(define-key jakarta-mode-map "\C-cc" 'jakarta-insert-record-form)
(define-key jakarta-mode-map "\C-cv" 'jakarta-insert-variable-form)
(define-key jakarta-mode-map "\C-cn" 'jakarta-insert-definition-form)
(define-key jakarta-mode-map "\C-cw" 'jakarta-insert-abstract-with-form)
(define-key jakarta-mode-map "\C-cf" 'jakarta-insert-function-form)
(define-key jakarta-mode-map "\C-ca" 'jakarta-insert-delete-form)
(define-key jakarta-mode-map "\C-cl" 'jakarta-insert-delete-in-form)
(define-key jakarta-mode-map "\C-co" 'jakarta-insert-drop-form)
(define-key jakarta-mode-map "\C-ce" 'jakarta-insert-coercion-form)
(define-key jakarta-mode-map "\C-cg" 'jakarta-insert-determine-form)
(define-key jakarta-mode-map "\C-ci" 'jakarta-insert-into-log-form)
(define-key jakarta-mode-map "\C-ct" 'jakarta-insert-totalize-form)
(define-key jakarta-mode-map "\C-cp" 'jakarta-insert-replace-by-in-form)
(define-key jakarta-mode-map "\C-cr" 'jakarta-insert-replace-by-form)
(define-key jakarta-mode-map "\C-cu" 'jakarta-insert-reduce-form)
(define-key jakarta-mode-map "\C-cx" 'jakarta-insert-select-in-form)
(define-key jakarta-mode-map "\C-cs" 'jakarta-insert-select-form)

(define-key jakarta-mode-map "\C-cm" 'jakarta-insert-remove-in-form)
(define-key jakarta-mode-map "\C-cj" 'jakarta-insert-reject-in-form)
(define-key jakarta-mode-map "\C-cy" 'jakarta-insert-replace_type-form)
(define-key jakarta-mode-map "\C-c\C-j" 'jakarta-jakartadoc-jakarta)

(defvar jakarta-mode-syntax-table (make-syntax-table)
  "Syntax table in use in Jakarta mode buffers.")

(modify-syntax-entry ?_ "_" jakarta-mode-syntax-table)
(modify-syntax-entry ?? "w" jakarta-mode-syntax-table)
(modify-syntax-entry ?~ "w" jakarta-mode-syntax-table)
(modify-syntax-entry ?: "." jakarta-mode-syntax-table)
(modify-syntax-entry ?' "w" jakarta-mode-syntax-table)
;; ' is part of words (for primes)
(if jakarta-support-jakartalight
    (modify-syntax-entry ?` "\"" jakarta-mode-syntax-table)
  (modify-syntax-entry ?` "." jakarta-mode-syntax-table))
;; ` is punctuation or character delimiter (Jakarta Light compatibility)
(modify-syntax-entry ?\" "\"" jakarta-mode-syntax-table)
;; " is a string delimiter
(modify-syntax-entry ?\\ "\\" jakarta-mode-syntax-table)
(modify-syntax-entry ?\( "()1" jakarta-mode-syntax-table)
(modify-syntax-entry ?*  ".23" jakarta-mode-syntax-table)
(modify-syntax-entry ?\) ")(4" jakarta-mode-syntax-table)
(let ((i 192))
  (while (< i 256)
    (modify-syntax-entry i "w" jakarta-mode-syntax-table)
    (setq i (1+ i))))

(defconst jakarta-font-lock-syntax
  '((?_ . "w") (?` . ".") (?\" . ".") (?\( . ".") (?\) . ".") (?* . "."))
  "Syntax changes for Font-Lock.")

(defvar jakarta-mode-abbrev-table ()
  "Abbrev table used for Jakarta mode buffers.")
(defun jakarta-define-abbrev (keyword)
  (define-abbrev jakarta-mode-abbrev-table keyword keyword 'jakarta-abbrev-hook))
(if jakarta-mode-abbrev-table ()
  (setq jakarta-mode-abbrev-table (make-abbrev-table))
  (mapcar 'jakarta-define-abbrev
	  '("module" "class" "object" "type" "val" "inherit" "virtual"
	    "constraint" "exception" "external" "open" "method" "and"
	    "initializer" "to" "downto" "do" "done" "else" "begin" "end"
	    "let" "in" "then" "with"))
  (setq abbrevs-changed nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                              The major mode

;;;###autoload
(defun jakarta-mode ()
  "Major mode for editing Jakarta code.

Dedicaced to Emacs and XEmacs, version 20 and higher. Provides
automatic indentation and compilation interface. Performs font/color
highlighting using Font-Lock. It is designed for Objective Jakarta but
handles Objective Labl and Jakarta Light as well.

Report bugs, remarks and questions to desousa@di.ubi.pt.

If you use XEmacs, the Font-Lock minor-mode is used accordingly to
your customization options. Within Emacs, you may want to use
Font-Lock in addding the following lines to the configuration file:

  (if (and (boundp 'window-system) window-system)
      (require 'font-lock))

Within XEmacs (more generally, if you have variable-sized fonts
and `atomic-extents' supported) you may also want to use Sym-Lock:

(if (and (boundp 'window-system) window-system)
    (when (string-match \"XEmacs\" emacs-version)
       	(if (not (and (boundp 'mule-x-win-initted) mule-x-win-initted))
            (require 'sym-lock))
       	(require 'font-lock)))

Emacs supports neither variable-sized fonts nor atomic sequences, perhaps
it is a good occasion for you to try XEmacs!

You have better byte-compile jakarta.el (and sym-lock.el if you use it)
because symbol highlighting is very time consuming.

For customization purposes, you should use `jakarta-mode-hook'
(run for every file) or `jakarta-load-hook' (run once) and not patch
the mode itself. You should add to your configuration file something like:
  (add-hook 'jakarta-mode-hook
            '(lambda ()
               ... ; your customization code
             ))
For example you can change the indentation of some keywords, the
`electric' flags, Font-Lock colors... Every customizable variable is
documented, use `C-h-v' or look at the mode's source code.

A special case is Sym-Lock customization: You may set
`jakarta-sym-lock-keywords' in your `.emacs' configuration file
to override default Sym-Lock patterns.

`custom-jakarta.el' is a sample customization file for standard changes.
You can append it to your `.emacs' or use it as a tutorial.

`M-x jakartadebug' FILE starts the Jakarta debugger jakartadebug on the executable
FILE, with input and output in an Emacs buffer named *jakartadebug-FILE*.

A Jakarta Interactive Mode to evaluate expressions in a toplevel is included.
Type `M-x jakarta-run-jakarta' or see special-keys below.

Some elementary rules have to be followed in order to get the best of
indentation facilities.
  - Because the `function' keyword has a special indentation (to handle
    case matchs) use the `fun' keyword when no case match is done.
  - Prefer the `or' keyword to `||' (they are semantically equivalent),
    it avoids some unwanted electric indentations.
  - In Jakarta, `;;' is no longer necessary for correct indentation,
    except before top level phrases not introduced by `type', `val', `let'
    etc. (i.e. phrases used for their side-effects or to be executed
    in a top level.)
  - Long sequences of `and's may slow down indentation slightly, since
    some computations (few actually) require to go back to the
    beginning of the sequence. Some very long nested blocks may also lead
    to slow processing of `end's, `else's, `done's...
  - Multiline strings are handled properly, but the string concatenation `^'
    is preferred to break long strings (the C-j keystroke can help in some
    cases).

Known bugs: (the same as ocaml mode, tuareg)
  - With Emacs: Text highlighting may be incorrect in some tricky examples
    (nested comments and strings), but there exists no proper solution
    to this without rewritting Font-Lock algorithms. This has been done
    with the improved XEmacs Font-Lock interface: Use XEmacs!
  - When writting a line with mixed code and comments, avoid putting
    comments at the beginning or middle of the text. More precisely, 
    writing comments immediately after `=' or parentheses then writing
    some more code on the line leads to indentation errors: You may write
    `let x (* blah *) = blah' but should avoid `let x = (* blah *) blah'.
    I'll try to fix this, even if I did not get any bug report about
    it so far...

Special keys for Jakarta mode:\\{jakarta-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'jakarta-mode)
  (setq mode-name "Jakarta")
  (use-local-map jakarta-mode-map)
  (set-syntax-table jakarta-mode-syntax-table)
  (setq local-abbrev-table jakarta-mode-abbrev-table)

  (if jakarta-window-system (jakarta-build-menu))

  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^[ \t]*$\\|\\*)$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "(* ")
  (make-local-variable 'comment-end)
  (setq comment-end " *)")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "(\\*+[ \t]*")
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line t)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'jakarta-indent-command)
  (make-variable-buffer-local 'before-change-functions)
  (add-hook 'before-change-functions 'jakarta-before-change-function)
  (add-hook 'before-change-functions 'jakarta-before-change-function)
  (make-variable-buffer-local 'normal-auto-fill-function)
  (setq normal-auto-fill-function 'jakarta-auto-fill-function)
	 
  ;; hooks for jakarta-mode, use them for jakarta-mode configuration
  (run-hooks 'jakarta-mode-hook)
  (jakarta-install-font-lock)
  (if jakarta-use-abbrev-mode (abbrev-mode 1))
  (message (concat "Major mode for Jakarta programs, "
		   jakarta-mode-version ".")))

(defun jakarta-install-font-lock (&optional no-sym-lock)
  (if (and (featurep 'font-lock)
	   (or jakarta-window-system jakarta-with-xemacs))
      (progn
	(jakarta-set-font-lock-faces)
	(if (and (not no-sym-lock)
		 (featurep 'sym-lock)) ; AFTER jakarta-set-font-lock-faces
	    (progn
	      (setq sym-lock-color
		    (face-foreground 'jakarta-font-lock-operator-face))
	      (if (not sym-lock-keywords)
		  (sym-lock jakarta-sym-lock-keywords))))
	(add-hook 'font-lock-mode-hook 'jakarta-font-lock-hook)
	(make-variable-buffer-local 'font-lock-defaults)
	(setq font-lock-defaults
	      (list 'jakarta-font-lock-keywords t nil
		    jakarta-font-lock-syntax nil
		    '(font-lock-fontify-buffer-function
		      . jakarta-fontify-buffer)
		    '(font-lock-fontify-region-function
		      . jakarta-fontify-region)))
	(make-variable-buffer-local 'font-lock-fontify-buffer-function)
	(if (boundp 'font-lock-fontify-buffer-function)
	    (setq font-lock-fontify-buffer-function 'jakarta-fontify-buffer)
	  (add-hook 'font-lock-after-fontify-buffer-hook
		    'jakarta-after-fontify-buffer))
	(make-variable-buffer-local 'font-lock-fontify-region-function)
	(if (boundp 'font-lock-fontify-region-function)
	    (setq font-lock-fontify-region-function 'jakarta-fontify-region))
	(font-lock-set-defaults)
	(if (not (or jakarta-with-xemacs font-lock-mode))
	    (font-lock-mode 1)) ; useful for beginners if not standard
	'font-lock)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Error processing

(require 'compile)

;; In some versions of Emacs, the regexps in
;; compilation-error-regexp-alist do not match the error messages when
;; the language is not English. Hence we add a regexp.

(defconst jakarta-error-regexp
  "^[A-\377]+ \"\\([^\"\n]+\\)\", [A-\377]+ \\([0-9]+\\)[-,:]"
  "Regular expression matching the error messages produced by (o)jakartac.")

(if (boundp 'compilation-error-regexp-alist)
    (or (assoc jakarta-error-regexp
               compilation-error-regexp-alist)
        (setq compilation-error-regexp-alist
              (cons (list jakarta-error-regexp 1 2)
               compilation-error-regexp-alist))))

;; A regexp to extract the range info.

(defconst jakarta-error-chars-regexp
  ".*, .*, [A-\377]+ \\([0-9]+\\)-\\([0-9]+\\):"
  "Regular expression extracting the character numbers
from an error message produced by (o)jakartac.")

;; Wrapper around next-error.

;; itz 04-21-96 somebody didn't get the documentation for next-error
;; right. When the optional argument is a number n, it should move
;; forward n errors, not reparse.

;; itz 04-21-96 instead of defining a new function, use defadvice
;; that way we get our effect even when we do \C-x` in compilation buffer  

(defadvice next-error (after jakarta-next-error activate)
 "Reads the extra positional information provided by the Jakarta compiler.

Puts the point and the mark exactly around the erroneous program
fragment. The erroneous fragment is also temporarily highlighted if
possible."
 (if (eq major-mode 'jakarta-mode)
     (let ((beg nil) (end nil))
       (save-excursion
	 (set-buffer compilation-last-buffer)
	 (save-excursion
	   (goto-char (window-point (get-buffer-window (current-buffer) t)))
	   (if (looking-at jakarta-error-chars-regexp)
	       (setq beg (string-to-int (jakarta-match-string 1))
		     end (string-to-int (jakarta-match-string 2))))))
       (beginning-of-line)
       (if beg
	   (progn
	     (setq beg (+ (point) beg) end (+ (point) end))
	     (goto-char beg) (push-mark end t t))))))

(defvar jakarta-interactive-error-regexp
  (concat "\\(\\("
	  "Toplevel input:"
	  "\\|Entr.e interactive:"
	  "\\|Characters [0-9-]*:"
	  "\\|The global value [^ ]* is referenced before being defined."
	  "\\|La valeur globale [^ ]* est utilis.e avant d'.tre d.finie."
	  "\\|Reference to undefined global"
	  "\\|The C primitive \"[^\"]*\" is not available."
	  "\\|La primitive C \"[^\"]*\" est inconnue."
	  "\\|Cannot find \\(the compiled interface \\)?file"
	  "\\|L'interface compil.e [^ ]* est introuvable."
	  "\\|Le fichier [^ ]* est introuvable."
	  "\\|Exception non rattrap.e:"
	  "\\|Uncaught exception:"
	  "\\)[^#]*\\)" )
  "Regular expression matching the error messages produced by Jakarta.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Indentation stuff

(defconst jakarta-keyword-regexp "\\<\\(object\\|initializer\\|and\\|c\\(onstraint\\|lass\\)\\|m\\(atch\\|odule\\|ethod\\|utable\\)\\|s\\(ig\\|truct\\)\\|begin\\|e\\(lse\\|x\\(ception\\|ternal\\)\\)\\|t\\(o\\|hen\\|ry\\|ype\\)\\|v\\(irtual\\|al\\)\\|w\\(h\\(ile\\|en\\)\\|ith\\)\\|i\\(f\\|n\\(herit\\)?\\)\\|f\\(or\\|un\\(ct\\(or\\|ion\\)\\)?\\)\\|let\\|do\\(wnto\\)?\\|parser?\\|rule\\|of\\)\\>\\|->\\|[;,|]"
  "Regexp for all recognized keywords.")

(defconst jakarta-match-|-keyword-regexp
  "\\<\\(and\\|fun\\(ction\\)?\\|type\\|with\\|parser?\\)\\>\\|[[({|=]"
  "Regexp for keywords supporting case match.")

(defconst jakarta-operator-regexp "[---+*/=<>@^&|]\\|:>\\|::\\|\\<\\(or\\|l\\(and\\|x?or\\|s[lr]\\)\\|as[lr]\\|mod\\)\\>"
  "Regexp for all operators.")

(defconst jakarta-kwop-regexp (concat jakarta-keyword-regexp "\\|=")
  "Regexp for all keywords, and the = operator which is generally
considered as a special keyword.")

(defconst jakarta-matching-keyword-regexp
  "\\<\\(and\\|do\\(ne\\)?\\|e\\(lse\\|nd\\)\\|in\\|then\\|\\(down\\)?to\\)\\>"
  "Regexp matching Jakarta keywords which act as end block delimiters.")

(defconst jakarta-leading-kwop-regexp
  (concat jakarta-matching-keyword-regexp "\\|\\<with\\>\\|[|>]?\\]\\|>?}\\|[|)]\\|.")
  "Regexp matching Jakarta keywords which need special indentation.")

(defconst jakarta-governing-phrase-regexp
  "\\<\\(val\\|type\\|m\\(ethod\\|odule\\)\\|c\\(onstraint\\|lass\\)\\|in\\(herit\\|itializer\\)\\|ex\\(ternal\\|ception\\)\\|open\\|let\\|object\\)\\>"
  "Regexp matching jakarta phrase delimitors.")

(defconst jakarta-governing-phrase-regexp-with-break
  (concat jakarta-governing-phrase-regexp "\\|."))

(defconst jakarta-keyword-alist
  '(("module" . jakarta-default-indent)
    ("class" . jakarta-class-indent)
    ("sig" . jakarta-sig-struct-indent)
    ("struct" . jakarta-sig-struct-indent)
    ("method" . jakarta-method-indent)
    ("object" . jakarta-begin-indent)
    ("begin" . jakarta-begin-indent)
    ("for" . jakarta-for-while-indent)
    ("while" . jakarta-for-while-indent)
    ("do" . jakarta-do-indent)
    ("type" . jakarta-type-indent) ; in some cases, `type' acts like a match
    ("val" . jakarta-val-indent)
    ("fun" . jakarta-fun-indent)
    ("if" . jakarta-if-then-else-indent)
    ("then" . jakarta-if-then-else-indent)
    ("else" . jakarta-if-then-else-indent)
    ("let" . jakarta-let-indent)
    ("match" . jakarta-match-indent)
    ("try" . jakarta-try-indent)
    ("rule" . jakarta-rule-indent)
    ("data" . jakarta-data-indent)

    ;; case match keywords
    ("function" . jakarta-function-indent)
    ("with" . jakarta-with-indent)
    ("parse" . jakarta-parse-indent)
    ("parser" . jakarta-parser-indent)

    ;; default indentation keywords
    ("when" . jakarta-default-indent)
    ("functor" . jakarta-default-indent)
    ("exception" . jakarta-default-indent)
    ("inherit" . jakarta-default-indent)
    ("initializer" . jakarta-default-indent)
    ("constraint" . jakarta-default-indent)
    ("virtual" . jakarta-default-indent)
    ("mutable" . jakarta-default-indent)
    ("external" . jakarta-default-indent)
    ("in" . jakarta-in-indent)
    ("of" . jakarta-default-indent)
    ("to" . jakarta-default-indent)
    ("downto" . jakarta-default-indent)
    ("[" . jakarta-default-indent)
    ("(" . jakarta-default-indent)
    ("{" . jakarta-default-indent)
    ("->" . jakarta-default-indent)
    ("|" . jakarta-default-indent))
"Association list of indentation values based on governing keywords.")

(defconst jakarta-leading-kwop-alist
  '(("|" . jakarta-find-|-match)
    ("}" . jakarta-find-match)
    (">}" . jakarta-find-match)
    (")" . jakarta-find-match)
    ("]" . jakarta-find-match)
    ("|]" . jakarta-find-match)
    (">]" . jakarta-find-match)
    ("end" . jakarta-find-match)
    ("done" . jakarta-find-done-match)
    ("in"  . jakarta-find-in-match)
    ("with" . jakarta-find-with-match)
    ("else" . jakarta-find-else-match)
    ("then" . jakarta-find-match)
    ("do" . jakarta-find-do-match)
    ("to" . jakarta-find-match)
    ("downto" . jakarta-find-match)
    ("and" . jakarta-find-and-match))
  "Association list used in Jakarta mode for skipping back over nested blocks.")

(defun jakarta-find-meaningful-word ()
  "Look back for a word, skipping comments and blanks.
Returns the actual text of the word, if found."
  (let ((found nil) (kwop nil))
    (while (and (not found)
		(re-search-backward
		 "[^ \t\n'_0-9A-Za-z\277-\377]\\|\\<\\(\\w\\|_\\)+\\>\\|\\*)"
		 (point-min) t))
      (setq kwop (jakarta-match-string 0))
      (if kwop
	  (if (jakarta-in-comment-p)
	      (jakarta-beginning-of-literal-or-comment-fast)
	    (setq found t))
	(setq found t)))
    (if found kwop (goto-char (point-min)) nil)))

(defconst jakarta-find-kwop-regexp
  (concat jakarta-matching-keyword-regexp "\\|\\<\\(data\\|for\\|while\\|do\\|if\\|begin\\|s\\(ig\\|truct\\)\\|object\\)\\>\\|[][(){}<>]\\|\\*)"))
(defun jakarta-make-find-kwop-regexp (kwop-regexp)
  (concat jakarta-find-kwop-regexp "\\|" kwop-regexp))

(defun jakarta-find-kwop (kwop-regexp &optional do-not-skip-regexp)
  "Look back for a Jakarta keyword or operator matching KWOP-REGEXP.
Skips blocks etc...

Ignore occurences inside literals and comments.
If found, return the actual text of the keyword or operator."
  (let ((found nil) (kwop nil))
    (while (and (not found)
		(re-search-backward kwop-regexp (point-min) t)
		(setq kwop (jakarta-match-string 0)))
      (cond
       ((jakarta-in-literal-or-comment-p)
	(jakarta-beginning-of-literal-or-comment-fast))
       ((looking-at "[]})>]")
	(jakarta-backward-up-list))
       ((jakarta-at-phrase-break-p)
	(setq found t))
       ((and do-not-skip-regexp (looking-at do-not-skip-regexp))
	(if (and (string= kwop "|") (char-equal ?| (preceding-char)))
	    (backward-char)
	  (setq found t)))
       ((looking-at jakarta-matching-keyword-regexp)
	(funcall (cdr (assoc (jakarta-match-string 0)
			     jakarta-leading-kwop-alist))))
       (t (setq found t))))
    (if found kwop (goto-char (point-min)) nil)))

(defun jakarta-find-match ()
  (jakarta-find-kwop jakarta-find-kwop-regexp))

(defconst jakarta-find-with-match-regexp
  (jakarta-make-find-kwop-regexp
   "\\<\\(match\\|try\\|module\\|begin\\)\\>\\|[[{(<]"))
(defun jakarta-find-with-match ()
  (jakarta-find-kwop jakarta-find-with-match-regexp))

(defconst jakarta-find-in-match-regexp
  (jakarta-make-find-kwop-regexp "\\<let\\>"))
(defun jakarta-find-in-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-in-match-regexp "\\<and\\>")))
    (cond ((string= kwop "and") (jakarta-find-in-match))
	  (t kwop))))

(defconst jakarta-find-else-match-regexp
  (jakarta-make-find-kwop-regexp ";"))
(defun jakarta-find-else-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-else-match-regexp
				     "\\<then\\>")))
    (cond ((string= kwop "then")
	   (jakarta-find-match) kwop)
	  ((string= kwop ";")
	   (jakarta-find-semi-colon-match)
	   (jakarta-find-else-match) kwop))))

(defun jakarta-find-do-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-kwop-regexp
				   "\\<\\(down\\)?to\\>")))
    (if (or (string= kwop "to") (string= kwop "downto"))
	(jakarta-find-match) kwop)))

(defun jakarta-find-done-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-kwop-regexp "\\<do\\>")))
    (if (string= kwop "do")
	(jakarta-find-do-match) kwop)))

(defconst jakarta-find-and-match-regexp
  "\\<\\(do\\(ne\\)?\\|e\\(lse\\|nd\\)\\|in\\|then\\|\\(down\\)?to\\)\\>\\|\\<\\(for\\|while\\|do\\|if\\|begin\\|s\\(ig\\|truct\\)\\|class\\)\\>\\|[][(){}]\\|\\*)\\|\\<\\(rule\\|exception\\|let\\|in\\|type\\|val\\|module\\)\\>")
(defconst jakarta-find-and-match-regexp-dnr
  (concat jakarta-find-and-match-regexp "\\|\\<and\\>"))
(defun jakarta-find-and-match (&optional do-not-recurse)
  (let* ((kwop (jakarta-find-kwop (if do-not-recurse
					  jakarta-find-and-match-regexp-dnr
					  jakarta-find-and-match-regexp)
				      "\\<and\\>"))
	 (old-point (point)))
    (cond ((or (string= kwop "type") (string= kwop "module"))
	   (let ((kwop2 (jakarta-find-meaningful-word)))
	     (cond ((string= kwop2 "with") kwop2)
		   ((string= kwop2 "and") (jakarta-find-and-match))
		   ((string= kwop "module") (string= kwop2 "let") kwop2)
		   (t (goto-char old-point) kwop))))
	  (t kwop))))

(defconst jakarta-find-=-match-regexp
  (jakarta-make-find-kwop-regexp "\\<\\(val\\|let\\|m\\(ethod\\|odule\\)\\|type\\|class\\|when\\|i[fn]\\)\\>\\|="))
(defun jakarta-find-=-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-=-match-regexp
				     "\\<\\(and\\|in\\)\\>\\|=")))
    (if (string= kwop "and")
	(jakarta-find-and-match)
      (if (and (string= kwop "=") (not (jakarta-false-=-p)))
	  (progn
	    (jakarta-find-=-match)
	    (jakarta-find-=-match))
	kwop))))

(defun jakarta-if-when-= ()
  (save-excursion
    (jakarta-find-=-match)
    (looking-at "\\<\\(if\\|when\\)\\>")))

(defun jakarta-captive-= ()
  (save-excursion
    (jakarta-find-=-match)
    (looking-at "\\<\\(if\\|when\\|module\\|type\\|class\\)\\>")))

(defconst jakarta-find-|-match-regexp
  (jakarta-make-find-kwop-regexp
   "\\<\\(with\\|fun\\(ction\\)?\\|type\\|parser?\\)\\>\\|[=|]"))
(defun jakarta-find-|-match ()
  (let* ((kwop (jakarta-find-kwop jakarta-find-|-match-regexp
				    "\\<\\(and\\|with\\)\\>\\||"))
	 (old-point (point)))
    (cond ((string= kwop "and")
	   (setq old-point (point))
	   (setq kwop (jakarta-find-and-match))
	   (goto-char old-point)
	   kwop)
	  ((and (string= kwop "|")
		(looking-at "|[^|]")
		(jakarta-in-indentation-p)) kwop)
	  ((string= kwop "|") (jakarta-find-|-match))
	  ((and (string= kwop "=")
		(or (looking-at "=[ \t]*\\((\\*\\|$\\)")
		    (jakarta-false-=-p)
		    (not (string= (save-excursion (jakarta-find-=-match))
				  "type"))))
	   (jakarta-find-|-match))
	  ((string= kwop "parse")
	   (if (and (string-match "\\.djl" (buffer-name))
		    (save-excursion
		      (string= (jakarta-find-meaningful-word) "=")))
	       kwop (jakarta-find-|-match)))
	  (t kwop))))

(defconst jakarta-find-->-match-regexp
  (jakarta-make-find-kwop-regexp "\\<\\(external\\|val\\|method\\|let\\|with\\|fun\\(ction\\)?\\|parser\\)\\>\\|[|:;]"))
(defun jakarta-find-->-match ()
  (let ((kwop (jakarta-find-kwop jakarta-find-->-match-regexp "\\<with\\>")))
    (cond
     ((string= kwop "|")
      (if (jakarta-in-indentation-p)
	  kwop
	(prog2 (forward-char -1) (jakarta-find-->-match))))
     ((not (string= kwop ":")) kwop)
     ;; if we get this far, we know we're looking at a colon.
     ((or (char-equal (char-before) ?:)
	  (char-equal (char-after (1+ (point))) ?:)
	  (char-equal (char-after (1+ (point))) ?>))
      (jakarta-find-->-match))
     ;; patch by T. Freeman
     (t (let ((oldpoint (point))
	      (match (jakarta-find-->-match)))
	  (if (looking-at ":")
	      match
	    (progn
	      ;; go back to where we were before the recursive call.
	      (goto-char oldpoint)
	      kwop)))))))

(defconst jakarta-find-semi-colon-match-regexp
  (jakarta-make-find-kwop-regexp ";[ \t]*\\((\\*\\|$\\)\\|->\\|\\<\\(let\\|method\\|with\\|try\\|initializer\\)\\>"))
(defun jakarta-find-semi-colon-match (&optional leading-semi-colon)
  (jakarta-find-kwop jakarta-find-semi-colon-match-regexp
			 "\\<\\(in\\|end\\|and\\|do\\|with\\)\\>")
  ;; we don't need to find the keyword matching `and' since we know it's `let'!
  (cond
   ((looking-at ";[ \t]*\\((\\*\\|$\\)")
    (forward-line 1)
    (while (or (jakarta-in-comment-p)
	       (looking-at "^[ \t]*\\((\\*\\|$\\)"))
      (forward-line 1))
    (back-to-indentation)
    (current-column))
   ((and leading-semi-colon
	 (looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*[^ \t\n]")
	 (not (looking-at "[[{(][|<]?[ \t]*\\((\\*\\|$\\)")))
    (current-column))
   ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
    (jakarta-back-to-paren-or-indentation t)
    (+ (current-column) jakarta-default-indent))
   ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*[^ \t\n]")
    (jakarta-search-forward-paren)
    (current-column))
   ((looking-at "\\<method\\>[ \t]*\\((\\*\\|$\\)")
    (jakarta-back-to-paren-or-indentation)
    (+ (current-column) jakarta-method-indent))
   ((looking-at "\\<begin\\>[ \t]*\\((\\*\\|$\\)")
    (jakarta-back-to-paren-or-indentation t)
    (+ (current-column) jakarta-begin-indent))
   ((looking-at "->")
    (if (save-excursion
	  (jakarta-find-->-match)
	  (looking-at "\\<\\(with\\|fun\\(ction\\)?\\|parser\\)\\>\\||"))
	(prog2
	    (jakarta-back-to-paren-or-indentation)
	    (+ (current-column) jakarta-default-indent))
      (jakarta-find-semi-colon-match)))
   ((looking-at "\\<end\\>")
    (jakarta-find-match)
    (jakarta-find-semi-colon-match))
   ((looking-at "\\<in\\>")
    (jakarta-find-in-match)
    (jakarta-back-to-paren-or-indentation)
    (+ (current-column) jakarta-in-indent))
   ((looking-at "\\<let\\>")
    (+ (current-column) jakarta-let-indent))
   (t (jakarta-back-to-paren-or-indentation t)
      (+ (current-column) jakarta-default-indent))))

(defconst jakarta-find-phrase-indentation-regexp
  (jakarta-make-find-kwop-regexp (concat jakarta-governing-phrase-regexp
					"\\|\\<and\\>")))
(defconst jakarta-find-phrase-indentation-regexp-pb
  (concat jakarta-find-phrase-indentation-regexp "\\|."))
(defconst jakarta-find-phrase-indentation-class-regexp
  (concat jakarta-matching-keyword-regexp "\\|\\<class\\>"))
(defun jakarta-find-phrase-indentation (&optional phrase-break)
  (if (and (looking-at "\\<\\(type\\|module\\)\\>") (> (point) (point-min))
	   (save-excursion
	     (jakarta-find-meaningful-word)
	     (looking-at "\\<\\(module\\|with\\|and\\|let\\)\\>")))
      (progn
	(jakarta-find-meaningful-word)
	(+ (current-column) jakarta-default-indent))
    (let ((looking-at-and (looking-at "\\<and\\>"))
	  (kwop (jakarta-find-kwop
		 (if phrase-break
		     jakarta-find-phrase-indentation-regexp-pb
		   jakarta-find-phrase-indentation-regexp)
		 "\\<\\(end\\|and\\|with\\|in\\)\\>"))
	  (tmpkwop nil) (curr nil))
      (if (and kwop (string= kwop "and"))
	  (setq kwop (jakarta-find-and-match)))
      (if (not kwop) (current-column)
	(cond
	 ((string= kwop "end")
	  (if (not (save-excursion
		     (setq tmpkwop (jakarta-find-match))
		     (setq curr (point))
		     (string= tmpkwop "object")))
	      (prog2
		  (jakarta-find-match)
		  (jakarta-find-phrase-indentation phrase-break))
	    (jakarta-find-kwop jakarta-find-phrase-indentation-class-regexp)
	    (current-column)))
	 ((and (string= kwop "with")
	       (not (save-excursion
		      (setq tmpkwop (jakarta-find-with-match))
		      (setq curr (point))
		      (string= tmpkwop "module"))))
	  (goto-char curr)
	  (jakarta-find-phrase-indentation phrase-break))
	 ((and (string= kwop "in")
	       (not (save-excursion
		      (setq tmpkwop (jakarta-find-in-match))
		      (if (string= tmpkwop "and")
			  (setq tmpkwop (jakarta-find-and-match)))
		      (setq curr (point))
		      (and (string= tmpkwop "let")
			   (not (jakarta-looking-at-expression-let))))))
	  (goto-char curr)
	  (jakarta-find-phrase-indentation phrase-break))
	 ((jakarta-at-phrase-break-p)
	  (end-of-line)
	  (jakarta-skip-blank-and-comments)
	  (current-column))
	 ((string= kwop "let")
	  (if (jakarta-looking-at-expression-let)
	      (jakarta-find-phrase-indentation phrase-break)
	    (current-column)))
	 ((string= kwop "with")
	  (current-column))
	 ((string= kwop "end")
	  (current-column))
	 ((string= kwop "in")
	  (jakarta-find-in-match)
	  (current-column))
	 ((string= kwop "class")
	  (jakarta-back-to-paren-or-indentation)
	  (current-column))
	 ((looking-at "\\<\\(object\\|s\\(ig\\|truct\\)\\)\\>")
	  (jakarta-back-to-paren-or-indentation t)
	  (+ (jakarta-assoc-indent kwop) (current-column)))
	 ((or (string= kwop "type") (string= kwop "module"))
	  (if (or (jakarta-looking-at-false-type)
		  (jakarta-looking-at-false-module))
	      (if looking-at-and (current-column)
		(jakarta-find-meaningful-word)
		(if (looking-at "\\<and\\>")
		    (prog2
			(jakarta-find-and-match)
			(jakarta-find-phrase-indentation phrase-break))
		  (jakarta-find-phrase-indentation phrase-break)))
	    (current-column)))
	 ((looking-at
	   "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
	  (jakarta-back-to-paren-or-indentation)
	  (+ (current-column) jakarta-default-indent))
	 ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*[^ \t\n]")
	  (jakarta-search-forward-paren)
	  (current-column))
	 ((string= kwop "open") ; compatible with Jakarta Light `#open'
	  (jakarta-back-to-paren-or-indentation) (current-column))
	 (t (current-column)))))))

(defconst jakarta-back-to-paren-or-indentation-regexp
  "[][(){}]\\|\\*)\\|^[ \t]*\\(.\\|\n\\)")
(defconst jakarta-back-to-paren-or-indentation-in-regexp
  (concat "\\<in\\>\\|" jakarta-back-to-paren-or-indentation-regexp))
(defconst jakarta-back-to-paren-or-indentation-lazy-regexp
  "[])}]\\|\\*)\\|^[ \t]*\\(.\\|\n\\)")
(defconst jakarta-back-to-paren-or-indentation-lazy-in-regexp
  (concat "\\<in\\>\\|" jakarta-back-to-paren-or-indentation-regexp))
(defun jakarta-back-to-paren-or-indentation (&optional forward-in)
  "Searches backwards for the first open paren in line, or skip to indentation.
Returns t iff skipped to indentation."
  (if (or (bolp) (jakarta-in-indentation-p)) (prog2 (back-to-indentation) t)
    (let ((kwop (jakarta-find-kwop
		 (if jakarta-lazy-paren
		     (if forward-in
			 jakarta-back-to-paren-or-indentation-lazy-in-regexp
		       jakarta-back-to-paren-or-indentation-lazy-regexp)
		   (if forward-in
		       jakarta-back-to-paren-or-indentation-in-regexp
		     jakarta-back-to-paren-or-indentation-regexp))
		 "\\<and\\|with\\|in\\>"))
	  (retval))
      (if (string= kwop "with")
	  (let ((with-point (point)))
	    (setq kwop (jakarta-find-with-match))
	    (if (or (string= kwop "match") (string= kwop "try"))
		(jakarta-find-kwop
		 jakarta-back-to-paren-or-indentation-regexp
		 "\\<and\\>")
	      (setq kwop "with") (goto-char with-point))))
      (setq retval
	    (cond
	     ((string= kwop "with") nil)
	     ((string= kwop "in") (jakarta-in-indentation-p))
	     ((looking-at "[[{(<]") (jakarta-search-forward-paren) nil)
	     (t (back-to-indentation) t)))
      (cond
       ((looking-at "|[^|]")
	(prog2 (re-search-forward "|[^|][ \t]*") nil))
       ((and forward-in (string= kwop "in"))
	(jakarta-find-in-match)
	(jakarta-back-to-paren-or-indentation forward-in)
	(if (looking-at "\\<\\(let\\|and\\)\\>")
	    (forward-char jakarta-in-indent)) nil)
       (t retval)))))

(defun jakarta-search-forward-paren ()
  (if jakarta-lazy-paren (jakarta-back-to-paren-or-indentation)
    (re-search-forward "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*")))

(defun jakarta-add-default-indent (leading-operator)
  (if leading-operator 0 jakarta-default-indent))

(defconst jakarta-compute-argument-indent-regexp
  (jakarta-make-find-kwop-regexp jakarta-kwop-regexp))
(defun jakarta-compute-argument-indent (leading-operator)
  (let ((old-point (save-excursion (beginning-of-line) (point)))
	(match-end-point) (kwop))
    (setq kwop (jakarta-find-kwop jakarta-compute-argument-indent-regexp
				 jakarta-keyword-regexp))
    (setq match-end-point (+ (point) (length kwop))) ; match-end is invalid !
    (cond
     ((string= kwop "->")
      (if (save-excursion (jakarta-find-->-match) (looking-at ":"))
	  (progn
	    (jakarta-find-->-match) ; slow, better to save the column
	    (jakarta-find-->-match) ; matching `val' or `let'
	    (+ (current-column) jakarta-val-indent))
	(jakarta-back-to-paren-or-indentation)
	(+ (jakarta-add-default-indent leading-operator) (current-column))))
     ((string= kwop "fun")
      (jakarta-back-to-paren-or-indentation t)
      (+ (current-column)
	 (jakarta-assoc-indent kwop)))
     ((and (if jakarta-lazy-= (looking-at "[[{(<]") (looking-at "[[<{(=]"))
	   (not (looking-at "=[ \t]*\\((\\*.*\\)?$")))
      (if (and (not jakarta-lazy-=)
	       (not (jakarta-if-when-=))
	       (looking-at "=")) (re-search-forward "=[ \t]*")
	(jakarta-search-forward-paren))
      (+ (jakarta-add-default-indent leading-operator)
	 (current-column)))
     ((<= old-point (point))
      (+ (jakarta-add-default-indent leading-operator) (current-column)))
     (t
      (forward-line 1)
      (beginning-of-line)
      (while (or (jakarta-in-comment-p) (looking-at "^[ \t]*\\((\\*.*\\)?$"))
	(forward-line 1))
      (jakarta-back-to-paren-or-indentation)
      (if (save-excursion (goto-char match-end-point)
			  (looking-at "[ \t]*\\((\\*.*\\)?$"))
	  (+ (jakarta-add-default-indent leading-operator)
	     (current-column))
	(current-column))))))

(defconst jakarta-compute-normal-indent-regexp
  (concat jakarta-compute-argument-indent-regexp "\\|^[ \t]*[^ \t\n]"))
(defun jakarta-compute-normal-indent ()
  (let ((leading-operator (looking-at jakarta-operator-regexp)))
    (beginning-of-line)
    (save-excursion
      (jakarta-find-meaningful-word)
      (if (looking-at jakarta-operator-regexp) (setq leading-operator t)))
    (save-excursion
      (let ((kwop (jakarta-find-kwop (if leading-operator
					jakarta-compute-argument-indent-regexp
				      jakarta-compute-normal-indent-regexp)
				    jakarta-keyword-regexp))
	    (leading-special 0))
	(if (string= kwop "and") (setq kwop (jakarta-find-and-match)))
	(while (and (string= kwop "=") (jakarta-false-=-p))
	  (setq kwop (jakarta-find-kwop jakarta-compute-normal-indent-regexp
				       jakarta-keyword-regexp))
	  (if (string= kwop "and") (setq kwop (jakarta-find-and-match))))
	(if (not kwop) (current-column)
	  (cond
	   ((jakarta-at-phrase-break-p)
	    (jakarta-find-phrase-indentation t))
	   ((and (string= kwop "|") (not  (char-equal ?\[ (preceding-char))))
	    (jakarta-backward-char)
	    (jakarta-back-to-paren-or-indentation)
	    (+ (current-column) jakarta-default-indent
	       (jakarta-add-default-indent leading-operator)))
	   ((or (looking-at "[[{(]")
		(and (looking-at "[<|]")
		     (char-equal ?\[ (preceding-char))
		     (prog2 (jakarta-backward-char) t))
		(and (looking-at "<")
		     (char-equal ?\{ (preceding-char))
		     (prog2 (jakarta-backward-char) t)))
	    (if (looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
		(prog2
		    (jakarta-back-to-paren-or-indentation t)
		    (+ jakarta-default-indent
		       (current-column))) ; parens do not operate
	      (jakarta-search-forward-paren)
	      (+ (jakarta-add-default-indent leading-operator)
		 (current-column))))
	   ((looking-at "->")
	    (let ((keyword-->-match (save-excursion (jakarta-find-->-match))))
	      (cond ((string= keyword-->-match "|")
		     (jakarta-find-->-match)
		     (re-search-forward "|[ \t]*")
		     (+ (current-column) jakarta-default-indent))
		    ((string= keyword-->-match ":")
		     (jakarta-find-->-match) ; slow, better to save the column
		     (jakarta-find-->-match) ; matching `val' or `let'
		     (+ (current-column) jakarta-val-indent))
		    (t (jakarta-back-to-paren-or-indentation)
		       (+ jakarta-default-indent (current-column))))))
	   ((looking-at jakarta-keyword-regexp)
	    (cond ((or (string= kwop ",") (string= kwop ";"))
		   (if (looking-at ";[ \t]*\\((\\*\\|$\\)")
		       (jakarta-find-semi-colon-match)
		     (if (looking-at ",[ \t]*\\((\\*\\|$\\)")
			 (progn
			   (jakarta-back-to-paren-or-indentation t)
			   (current-column))
		       (jakarta-back-to-paren-or-indentation t)
		       (+ (current-column) jakarta-default-indent))))
		  ((and (looking-at "\\<\\(in\\|begin\\|do\\)\\>\\|->")
			(not (looking-at
			      "\\([a-z]+\\|->\\)[ \t]*\\((\\*\\|$\\)")))
		   (if (string= kwop "in")
		       (re-search-forward "\\<in\\>[ \t]*")
		     (jakarta-back-to-paren-or-indentation t))
		   (+ (current-column)
		      (jakarta-add-default-indent leading-operator)
		      (if (string= kwop "in") 0 ; aligned, do not indent
			(jakarta-assoc-indent kwop))))
		  ((string= kwop "with")
		   (if (save-excursion
			 (let ((tmpkwop (jakarta-find-with-match)))
			   (or (string= tmpkwop "module")
			       (string= tmpkwop "{"))))
		       (prog2
			   (jakarta-back-to-paren-or-indentation)
			   (+ (current-column) jakarta-default-indent))
		     (jakarta-back-to-paren-or-indentation)
		     (+ (current-column)
			(jakarta-assoc-indent kwop t))))
		  ((string= kwop "in")
		   (jakarta-find-in-match)
		   (jakarta-back-to-paren-or-indentation)
		   (+ (current-column) jakarta-in-indent))
		  (t (jakarta-back-to-paren-or-indentation t)
		     (+ (current-column)
			(jakarta-assoc-indent kwop t)))))
	   ((and (looking-at "=") (not (jakarta-false-=-p)))
	    (if (or jakarta-lazy-= (jakarta-captive-=)
		    (looking-at "=[ \t]*\\((\\*\\|$\\)")) ; not perfect...
		(let ((current-column-module-type nil))
		  (+
		   (progn
		     (jakarta-find-=-match)
		     (save-excursion
		       (if (looking-at "\\<and\\>") (jakarta-find-and-match))
		       (cond
			((looking-at "\\<type\\>")
			 (jakarta-find-meaningful-word)
			 (if (looking-at "\\<module\\>")
			     (progn
			       (setq current-column-module-type
				     (current-column))
			       jakarta-default-indent)
			   (if (looking-at "\\<\\(with\\|and\\)\\>")
			       (progn
				 (setq current-column-module-type
				       (current-column))
				 jakarta-default-indent)
			     (re-search-forward "\\<type\\>")
			     (beginning-of-line)
			     (+ jakarta-type-indent
				jakarta-|-extra-unindent))))
			((looking-at
			  "\\<\\(val\\|let\\|m\\(ethod\\|odule\\)\\|class\\|when\\|\\|for\\|if\\)\\>")
			 (let ((matched-string (jakarta-match-string 0)))
			   (jakarta-back-to-paren-or-indentation t)
			   (setq current-column-module-type (current-column))
			   (jakarta-assoc-indent matched-string)))
			((looking-at "\\<object\\>")
			 (jakarta-back-to-paren-or-indentation t)
			 (setq current-column-module-type (current-column))
			 (+ (jakarta-assoc-indent "object")
			    jakarta-default-indent))
			(t (jakarta-search-forward-paren)
			   (setq current-column-module-type (current-column))
			   jakarta-default-indent))))
		   (if current-column-module-type
		       current-column-module-type
		     (current-column))))
	      (if (save-excursion (jakarta-find-=-match)
				  (looking-at "\\<sig\\>"))
		  (prog2
		      (jakarta-back-to-paren-or-indentation t)
		      (jakarta-assoc-indent "sig"))
		(+ (jakarta-add-default-indent leading-operator)
		   (prog2
		       (re-search-forward "=[ \t]*")
		       (current-column))))))
	   (nil 0)
	   (t (jakarta-compute-argument-indent leading-operator))))))))

(defun jakarta-looking-at-expression-let ()
  (save-excursion
    (and (jakarta-find-meaningful-word)
	 (not (jakarta-at-phrase-break-p))
	 (or (looking-at "[[({;=]\\|\\<\\(data\\)\\>")
	     (looking-at jakarta-operator-regexp)))))

(defun jakarta-looking-at-false-module ()
  (save-excursion (jakarta-find-meaningful-word)
		  (looking-at "\\<\\(let\\|with\\|and\\)\\>")))

(defun jakarta-looking-at-false-sig-struct ()
  (save-excursion (jakarta-find-module)
		  (looking-at "\\<module\\>")))

(defun jakarta-looking-at-false-type ()
  (save-excursion (jakarta-find-meaningful-word)
		  (looking-at "\\<\\(type\\|and\\)\\>")))

(defun jakarta-looking-at-in-let ()
  (save-excursion (string= (jakarta-find-meaningful-word) "in")))

(defconst jakarta-find-module-regexp
  (jakarta-make-find-kwop-regexp "\\<module\\>"))
(defun jakarta-find-module ()
  (jakarta-find-kwop jakarta-find-module-regexp))

(defun jakarta-modify-syntax ()
  "Switch to modified internal syntax."
  (modify-syntax-entry ?. "w" jakarta-mode-syntax-table)
  (modify-syntax-entry ?_ "w" jakarta-mode-syntax-table))

(defun jakarta-restore-syntax ()
  "Switch back to interactive syntax."
  (modify-syntax-entry ?. "." jakarta-mode-syntax-table)
  (modify-syntax-entry ?_ "_" jakarta-mode-syntax-table))

(defun jakarta-indent-command (&optional from-leading-star)
  "Indent the current line in Jakarta mode.

Compute new indentation based on Jakarta syntax."
  (interactive "*")
  (let ((old-cfs case-fold-search))
    (if (not from-leading-star)
	(jakarta-auto-fill-insert-leading-star))
    (setq case-fold-search nil)
    (jakarta-modify-syntax)
    (save-excursion
      (back-to-indentation)
      (indent-line-to (jakarta-compute-indent)))
    (if (jakarta-in-indentation-p) (back-to-indentation))
    (setq case-fold-search old-cfs)
    (jakarta-restore-syntax)))

(defun jakarta-compute-indent ()
  (save-excursion
    (cond
     ((jakarta-in-comment-p)
      (cond
       ((looking-at "(\\*")
	(if jakarta-indent-leading-comments
	    (save-excursion
	      (while (and (progn (beginning-of-line)
				 (> (point) 1))
			  (progn (forward-line -1)
				 (back-to-indentation)
				 (jakarta-in-comment-p))))
	      (if (looking-at "[ \t]*$")
		  (progn
		    (jakarta-skip-blank-and-comments)
		    (if (or (looking-at "$") (jakarta-in-comment-p))
			0
		      (jakarta-compute-indent)))
		(forward-line 1)
		(jakarta-compute-normal-indent)))
	  (current-column)))
       ((looking-at "\\*\\**)")
	(jakarta-beginning-of-literal-or-comment-fast)
	(if (jakarta-leading-star-p)
	    (+ (current-column)
	       (if (save-excursion
		     (forward-line 1)
		     (back-to-indentation)
		     (looking-at "*")) 1
		 jakarta-comment-end-extra-indent))
	  (+ (current-column) jakarta-comment-end-extra-indent)))
       (jakarta-indent-comments
	(let ((star (and (jakarta-leading-star-p)
			 (looking-at "\\*"))))
	  (jakarta-beginning-of-literal-or-comment-fast)
	  (if star (re-search-forward "(") (re-search-forward "(\\*[ \t]*"))
	  (current-column)))))
     ((jakarta-in-literal-p)
      (current-column))
     ((looking-at "\\<let\\>")
      (if (jakarta-looking-at-expression-let)
	  (if (jakarta-looking-at-in-let)
	      (progn
		(jakarta-find-meaningful-word)
		(jakarta-find-in-match)
		(jakarta-back-to-paren-or-indentation)
		(current-column))
	    (jakarta-compute-normal-indent))
	(jakarta-find-phrase-indentation)))
     ((looking-at jakarta-governing-phrase-regexp-with-break)
      (jakarta-find-phrase-indentation))
     ((and jakarta-sig-struct-align (looking-at "\\<\\(sig\\|struct\\)\\>"))
      (if (string= (jakarta-find-module) "module") (current-column)
	(jakarta-back-to-paren-or-indentation)
	(+ jakarta-default-indent (current-column))))
     ((looking-at ";") (jakarta-find-semi-colon-match t))
     ((or (looking-at "%\\|.")
	  (and jakarta-support-jakartalight (looking-at "#"))
	  (looking-at "#open")) 0)
     ((looking-at jakarta-leading-kwop-regexp)
      (let ((kwop (jakarta-match-string 0)))
	(let* ((old-point (point))
	       (paren-match-p (looking-at "[|>]?[]})]"))
	       (need-not-back-kwop (string= kwop "and"))
	       (real-| (looking-at "|\\([^|]\\|$\\)"))
	       (matching-kwop
		(if (string= kwop "and") (jakarta-find-and-match t)
		  (funcall (cdr (assoc kwop jakarta-leading-kwop-alist)))))
	       (match-|-keyword-p
		(looking-at jakarta-match-|-keyword-regexp)))
	  (cond
	   ((and (looking-at "\\(\\[|?\\|{<?\\|(\\)[ \t]*[^ \t\n]")
		 (not (looking-at "[[{(][|<]?[ \t]*\\((\\*\\|$\\)")))
	    (if (and (string= kwop "|") real-|) (current-column)
	      (if (not paren-match-p) (jakarta-search-forward-paren))
	      (if jakarta-lazy-paren (jakarta-back-to-paren-or-indentation))
	      (current-column)))
	   ((and (string= kwop "|") real-|)
	    (cond
	     ((string= matching-kwop "|")
	      (if (not need-not-back-kwop)
		  (jakarta-back-to-paren-or-indentation))
	      (current-column))
	     ((and (string= matching-kwop "=")
		   (not (jakarta-false-=-p)))
	      (re-search-forward "=[ \t]*")
	      (current-column))
	     (match-|-keyword-p
	      (if (not need-not-back-kwop)
		  (jakarta-back-to-paren-or-indentation))
	      (- (+ (jakarta-assoc-indent
		     matching-kwop t)
		    (current-column))
		 (if (string= matching-kwop "type") 0
		   jakarta-|-extra-unindent)))
	     (t (goto-char old-point)
		(jakarta-compute-normal-indent))))
	   ((and (string= kwop "|") (not real-|))
	    (goto-char old-point)
	    (jakarta-compute-normal-indent))
	   ((not need-not-back-kwop)
	    (jakarta-back-to-paren-or-indentation (not (string= kwop "in")))
	    (current-column))
	   (t (current-column))))))
     (t (jakarta-compute-normal-indent)))))

(defun jakarta-split-string ()
  "Called whenever a line is broken inside a Jakarta string literal."
  (insert-before-markers "\" ^\"")
  (jakarta-backward-char))

(defadvice newline-and-indent (around
			       jakarta-newline-and-indent
			       activate)
  "Handle multi-line strings in Jakarta mode."
    (let ((hooked (and (eq major-mode 'jakarta-mode) (jakarta-in-literal-p)))
	  (split-mark))
      (if (not hooked) nil
	(setq split-mark (set-marker (make-marker) (point)))
	(jakarta-split-string))
      ad-do-it
      (if (not hooked) nil
	(goto-char split-mark)
	(set-marker split-mark nil))))

(defun jakarta-electric ()
  "If inserting a | operator at beginning of line, reindent the line."
  (interactive "*")
  (let ((electric (and jakarta-electric-indent
		       (jakarta-in-indentation-p)
		       (not (jakarta-in-literal-p))
		       (not (jakarta-in-comment-p)))))
    (self-insert-command 1)
    (if (and electric
	     (not (and (char-equal ?| (preceding-char))
		       (save-excursion
			 (jakarta-backward-char)
			 (jakarta-find-|-match)
			 (not (looking-at jakarta-match-|-keyword-regexp))))))
	(jakarta-indent-command))))

(defun jakarta-electric-rp ()
  "If inserting a ) operator or a comment-end at beginning of line,
reindent the line."
  (interactive "*")
  (let ((electric (and jakarta-electric-indent
		       (or (jakarta-in-indentation-p)
			   (char-equal ?* (preceding-char)))
		       (not (jakarta-in-literal-p))
		       (or (not (jakarta-in-comment-p))
			   (save-excursion
			     (back-to-indentation)
			     (looking-at "\\*"))))))
    (self-insert-command 1)
    (if electric
	(jakarta-indent-command))))

(defun jakarta-electric-rc ()
  "If inserting a } operator at beginning of line, reindent the line.

Reindent also if } is inserted after a > operator at beginning of line.
Also, if the matching { is followed by a < and this } is not preceded
by >, insert one >."
  (interactive "*")
  (let* ((prec (preceding-char))
	 (look-bra (and jakarta-electric-close-vector
			(not (jakarta-in-literal-or-comment-p))
			(not (char-equal ?> prec))))
	 (electric (and jakarta-electric-indent
			(or (jakarta-in-indentation-p)
			    (and (char-equal ?> prec)
				 (save-excursion (jakarta-backward-char)
						 (jakarta-in-indentation-p))))
			(not (jakarta-in-literal-or-comment-p)))))
    (self-insert-command 1)
    (if look-bra
	(save-excursion
	  (let ((inserted-char
		 (save-excursion
		   (jakarta-backward-char)
		   (jakarta-backward-up-list)
		   (cond ((looking-at "{<") ">")
			 (t "")))))
	    (jakarta-backward-char)
	    (insert inserted-char))))
    (if electric (jakarta-indent-command))))

(defun jakarta-electric-rb ()
  "If inserting a ] operator at beginning of line, reindent the line.

Reindent also if ] is inserted after a | operator at beginning of line.
Also, if the matching [ is followed by a | and this ] is not preceded
by |, insert one |."
  (interactive "*")
  (let* ((prec (preceding-char))
	 (look-|-or-bra (and jakarta-electric-close-vector
			     (not (jakarta-in-literal-or-comment-p))
			     (not (and (char-equal ?| prec)
				       (not (char-equal
					     (save-excursion
					       (jakarta-backward-char)
					       (preceding-char)) ?\[))))))
	 (electric (and jakarta-electric-indent
			(or (jakarta-in-indentation-p)
			    (and (char-equal ?| prec)
				 (save-excursion (jakarta-backward-char)
						 (jakarta-in-indentation-p))))
			(not (jakarta-in-literal-or-comment-p)))))
    (self-insert-command 1)
    (if look-|-or-bra
	(save-excursion
	  (let ((inserted-char
		 (save-excursion
		   (jakarta-backward-char)
		   (jakarta-backward-up-list)
		   (cond ((looking-at "\\[|") "|")
			 (t "")))))
	    (jakarta-backward-char)
	    (insert inserted-char))))
    (if electric (jakarta-indent-command))))

(defun jakarta-abbrev-hook ()
  "If inserting a leading keyword at beginning of line, reindent the line."
  (if (not (jakarta-in-literal-or-comment-p))
      (let* ((bol (save-excursion (beginning-of-line) (point)))
	     (kw (save-excursion
		   (and (re-search-backward "^[ \t]*\\(\\w\\|_\\)+\\=" bol t)
			(jakarta-match-string 1)))))
	(if kw (progn
		   (insert " ")
		   (jakarta-indent-command)
		   (backward-delete-char-untabify 1))))))

(defun jakarta-skip-to-end-of-phrase ()
  (let ((old-point (point)))
    (if (and (string= (jakarta-find-meaningful-word) ".")
	     (char-equal (preceding-char) ?\.))
	(setq old-point (1- (point))))
    (goto-char old-point)
    (let ((kwop (jakarta-find-meaningful-word)))
      (goto-char (+ (point) (length kwop))))))

(defun jakarta-skip-blank-and-comments ()
  (skip-chars-forward " \t\n")
  (while (and (< (point) (point-max)) (jakarta-in-comment-p)
	      (re-search-forward "\\*)" (point-max) t))
    (skip-chars-forward " \t\n")))

(defun jakarta-skip-back-blank-and-comments ()
  (skip-chars-backward " \t\n")
  (while (save-excursion (jakarta-backward-char)
			 (and (> (point) (point-min)) (jakarta-in-comment-p)))
    (jakarta-backward-char)
    (jakarta-beginning-of-literal-or-comment) (skip-chars-backward " \t\n")))

(defconst jakarta-beginning-phrase-regexp
  "^#[ \t]*[a-z][_a-z]*\\>\\|\\<\\(and\\|abstract\\|data\\|type\\|module\\|function\\|record\\|declaration\\|variable\\|definition\\|into\\|build\\)\\>"
  "Regexp matching jakarta phrase delimitors.")
(defun jakarta-find-phrase-beginning ()
  "Find `real' phrase beginning and returns point."
  (beginning-of-line)
  (jakarta-skip-blank-and-comments)
  (end-of-line)
  (jakarta-skip-to-end-of-phrase)
  (let ((old-point (point)))
    (jakarta-find-kwop jakarta-beginning-phrase-regexp)
    (while (and (> (point) (point-min)) (< (point) old-point)
		(or (not (looking-at jakarta-beginning-phrase-regexp))
		    (and (looking-at "\\<data\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<variable\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<declaration\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<definition\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<into\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<build\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<abstract\\>")
			 (jakarta-looking-at-expression-let))
		    (and (looking-at "\\<module\\>")
			 (jakarta-looking-at-false-module))
		    (and (looking-at "\\<\\(record\\)\\>")
			 (jakarta-looking-at-false-sig-struct))
		    (and (looking-at "\\<type\\>")
			 (jakarta-looking-at-false-type))))
      (if (looking-at "\\<and\\>")
	  ;;(jakarta-find-match)
	  (jakarta-data-indent)
	(if (not (bolp)) (jakarta-backward-char))
	(setq old-point (point))
	(jakarta-find-kwop jakarta-beginning-phrase-regexp))
)
    (if (jakarta-at-phrase-break-p)
	(prog2 (end-of-line) (jakarta-skip-blank-and-comments)))
    (back-to-indentation)
    (point)))

(defconst jakarta-inside-block-opening "\\<\\(struct\\|sig\\|object\\)\\>")
(defun jakarta-search-forward-end-iter (begin)
  (if (re-search-forward "\\<end\\>" (point-max) t)
      (if (save-excursion
	    (jakarta-backward-char 3)
	    (jakarta-find-match)
	    (if (looking-at jakarta-inside-block-opening)
		(jakarta-find-phrase-beginning))
	    (> (point) begin))
	  (jakarta-search-forward-end-iter begin)
	t)
    nil))
(defun jakarta-search-forward-end ()
  (jakarta-search-forward-end-iter (point)))

(defconst jakarta-inside-block-opening-full
  (concat jakarta-inside-block-opening "\\|\\<\\(module\\|class\\)\\>"))
(defconst jakarta-inside-block-regexp
  (concat jakarta-matching-keyword-regexp "\\|" jakarta-inside-block-opening))
(defun jakarta-inside-block-find-kwop ()
  (let ((kwop (jakarta-find-kwop jakarta-inside-block-regexp
				"\\<\\(and\\|end\\)\\>")))
    (if (string= kwop "and") (setq kwop (jakarta-find-and-match)))
    (if (string= kwop "with") (setq kwop nil))
    (if (string= kwop "end")
	(progn
	  (jakarta-find-match)
	  (jakarta-find-kwop jakarta-inside-block-regexp)
	  (jakarta-inside-block-find-kwop))
      kwop)))
(defun jakarta-inside-block-p ()
  (let ((begin) (end) (and-end) (kwop t))
    (save-excursion
      (if (looking-at "\\<and\\>")
	  (jakarta-find-and-match))
      (setq begin (point))
      (if (or (and (looking-at "\\<class\\>")
		   (save-excursion
		     (re-search-forward "\\<object\\>"
					(point-max) t)
		     (jakarta-find-phrase-beginning)
		     (> (point) begin)))
	      (and (looking-at "\\<module\\>")
		   (save-excursion
		     (re-search-forward "\\<\\(sig\\|struct\\)\\>"
					(point-max) t)
		     (jakarta-find-phrase-beginning)
		     (> (point) begin)))) ()
	(if (not (looking-at jakarta-inside-block-opening-full))
	    (setq kwop (jakarta-inside-block-find-kwop)))
	(if (not kwop) ()
	  (setq begin (point))
	  (if (not (jakarta-search-forward-end)) ()
	    (jakarta-backward-char 3)
	    (if (not (looking-at "\\<end\\>")) ()
	      (jakarta-forward-char 3)
	      (setq end (point))
	      (setq and-end (point))
	      (jakarta-skip-blank-and-comments)
	      (while (looking-at "\\<and\\>")
		(setq and-end (point))
		(if (not (jakarta-search-forward-end)) ()
		  (jakarta-backward-char 3)
		  (if (not (looking-at "\\<end\\>")) ()
		    (jakarta-forward-char 3)
		    (setq and-end (point))
		    (jakarta-skip-blank-and-comments))))
	      (list begin end and-end))))))))

(defun jakarta-discover-phrase (&optional quiet)
  (end-of-line)
  (let ((end (point)) (old-cfs case-fold-search))
    (setq case-fold-search nil)
    (jakarta-modify-syntax)
    (jakarta-find-phrase-beginning)
    (if (> (point) end) (setq end (point)))
    (save-excursion
      (let ((begin (point)) (cpt 0) (lines-left 0) (stop)
	    (inside-block (jakarta-inside-block-p))
	    (looking-block (looking-at jakarta-inside-block-opening-full)))
	(if (and looking-block inside-block)
	    (progn
	      (setq begin (nth 0 inside-block))
	      (setq end (nth 2 inside-block))
	      (goto-char end))
	  (if inside-block
	      (progn
		(setq stop (save-excursion (goto-char (nth 1 inside-block))
					   (beginning-of-line) (point)))
		(if (< stop end) (setq stop (point-max))))
	    (setq stop (point-max)))
	  (save-restriction
	    (goto-char end)
	    (while (and (= lines-left 0)
			(or (not inside-block) (< (point) stop))
			(<= (save-excursion
			      (jakarta-find-phrase-beginning)) end))
	      (if (not quiet)
		  (prog2
		      (setq cpt (1+ cpt))
		      (if (= 8 cpt)
			  (message "Looking for enclosing phrase..."))))
	      (setq end (point))
	      (jakarta-skip-to-end-of-phrase)
	      (beginning-of-line)
	      (narrow-to-region (point) (point-max))
	      (goto-char end)
	      (setq lines-left (forward-line 1)))))
	(if (>= cpt 8) (message "Looking for enclosing phrase... done."))
	(save-excursion (jakarta-skip-blank-and-comments) (setq end (point)))
	(jakarta-skip-back-blank-and-comments)
	(setq case-fold-search old-cfs)
	(jakarta-restore-syntax)
	(list begin (point) end)))))

(defun jakarta-mark-phrase ()
  "Put mark at end of this Jakarta phrase, point at beginning.
The Jakarta phrase is the phrase just before the point."
  (interactive)
  (let ((pair (jakarta-discover-phrase)))
    (goto-char (nth 1 pair)) (push-mark (nth 0 pair) t t)))

(defun jakarta-next-phrase (&optional quiet)
  "Skip to the beginning of the next phrase."
  (interactive "i")
  (goto-char (save-excursion (nth 2 (jakarta-discover-phrase quiet))))
  (if (looking-at "\\<end\\>") (jakarta-next-phrase quiet))
  (if (looking-at ".")
      (progn
	(forward-char 2)
	(jakarta-skip-blank-and-comments))))

(defun jakarta-previous-phrase ()
  "Skip to the beginning of the previous phrase."
  (interactive)
  (beginning-of-line)
  (jakarta-skip-to-end-of-phrase)
  (jakarta-discover-phrase))

(defun jakarta-indent-phrase ()
  "Depending of the context: justify and indent a comment,
or indent all lines in the current phrase."
  (interactive)
  (save-excursion
    (back-to-indentation)
    (if (jakarta-in-comment-p)
	(let* ((cobpoint (save-excursion
			   (jakarta-beginning-of-literal-or-comment)
			   (point)))
	       (begpoint (save-excursion
			   (while (and (> (point) cobpoint)
				       (jakarta-in-comment-p)
				       (not (looking-at "^[ \t]*$")))
			     (forward-line -1))
			   (max cobpoint (point))))
	       (coepoint (save-excursion
			   (while (jakarta-in-comment-p)
			     (re-search-forward "\\*)"))
			   (point)))
	       (endpoint (save-excursion
			   (re-search-forward "^[ \t]*$" coepoint 'end)
			   (beginning-of-line)
			   (forward-line 1)
			   (point)))
	       (leading-star (jakarta-leading-star-p)))
	  (goto-char begpoint)
	  (while (and leading-star
		      (< (point) endpoint)
		      (not (looking-at "^[ \t]*$")))
	    (forward-line 1)
	    (back-to-indentation)
	    (if (looking-at "\\*\\**\\([^)]\\|$\\)")
		(progn
		  (delete-char 1)
		  (setq endpoint (1- endpoint)))))
	  (goto-char (min (point) endpoint))
	  (fill-region begpoint endpoint)
	  (re-search-forward "\\*)")
	  (setq endpoint (point))
	  (if leading-star
	      (progn
		(goto-char begpoint)
		(forward-line 1)
		(if (< (point) endpoint)
		    (jakarta-auto-fill-insert-leading-star t))))
	  (indent-region begpoint endpoint nil))
      (let ((pair (jakarta-discover-phrase)))
	(indent-region (nth 0 pair) (nth 1 pair) nil)))))

;; Auxiliary function (from J. Garrigue)
(defun jakarta-find-alternate-file ()
  "Switch Implementation/Interface."
  (interactive)
  (let ((name (buffer-file-name)))
    (if (string-match "^\\(.*\\)\\.\\(dj\\|dj\\)$" name)
	(find-file (concat (jakarta-match-string 1 name)
			   (if (string= "dj" (jakarta-match-string 2 name))
			       ".dj" ".dj"))))))
(defun jakarta-insert-type-form ()
  "Inserts a nicely formatted type form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "type   .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)  
    (beginning-of-line)
    (backward-char 4) 
    (jakarta-indent-command)))

(defun jakarta-insert-data-form ()
  "Inserts a nicely formatted data form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "data   =   .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 8) 
    (jakarta-indent-command)))


(defun jakarta-insert-record-form ()
  "Inserts a nicely formatted data form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "record   =  .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 7) 
    (jakarta-indent-command)))





(defun jakarta-insert-function-form ()
  "Inserts a nicely formatted function form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "function   :  ->  :=\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 12) 
    (jakarta-indent-command)))

(defun jakarta-insert-variable-form ()
  "Inserts a nicely formatted data form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "variable   :  .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 8) 
    (jakarta-indent-command)))

(defun jakarta-insert-definition-form ()
  "Inserts a nicely formatted data form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "definition   :  :=  .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 10) 
    (jakarta-indent-command)))

(defun jakarta-insert-build-from-form ()
  "Inserts a nicely formatted build from abstraction functions form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "build    from\nabstraction functions   .\n  ")
    (end-of-line)
    (indent-region old (point) nil)
    (push-mark)
    (forward-line -1)
    (jakarta-indent-command)
    (beginning-of-line 1)
    (backward-char 7)))

(defun jakarta-insert-abstract-with-form ()
  "Inserts a nicely formatted data form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "abstract   with  .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 10) 
    (jakarta-indent-command)))

(defun jakarta-insert-delete-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "delete \n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))


(defun jakarta-insert-delete-in-form ()
  "Inserts a nicely formatted select in form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "delete   in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 5)
    (jakarta-indent-command)))


(defun jakarta-insert-drop-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "drop \n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))


(defun jakarta-insert-coercion-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "coercion \n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))

(defun jakarta-insert-replace-by-in-form ()
  "Inserts a nicely formatted replace form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "replace   by\n  in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (forward-line -1)
    (jakarta-indent-command)
    (beginning-of-line 1)
    (backward-char 4)))


(defun jakarta-insert-replace-by-form ()
  "Inserts a nicely formatted select in form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "replace   by\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 5)
    (jakarta-indent-command)))
    



(defun jakarta-insert-reduce-form ()
  "Inserts a nicely formatted reduce form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "reduce   at\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 4)
    (jakarta-indent-command)))

(defun jakarta-insert-select-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "select \n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))

(defun jakarta-insert-select-in-form ()
  "Inserts a nicely formatted select in form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "select   in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 5)
    (jakarta-indent-command)))

(defun jakarta-insert-determine-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "determine \n ")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))

(defun jakarta-insert-into-log-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "into   log  .\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 9) 
    (jakarta-indent-command)))

(defun jakarta-insert-totalize-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "totalize \n ")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))


(defun jakarta-insert-remove-in-form ()
  "Inserts a nicely formatted select in form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "remove   in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 5)
    (jakarta-indent-command)))

(defun jakarta-insert-reject-in-form ()
  "Inserts a nicely formatted select in form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "reject   in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 5)
    (jakarta-indent-command)))

(defun jakarta-insert-replace_type-form ()
  "Inserts a nicely formatted select form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "replace_type  \n")
    (end-of-line)
    (indent-region old (point) nil)
    (jakarta-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 1) 
    (jakarta-indent-command)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            Jakarta interactive mode

;; Augment Jakarta mode, so you can process Jakarta code in the source files.

(require 'comint)

(defvar jakarta-interactive-mode-map nil)
(if jakarta-interactive-mode-map nil
  (setq jakarta-interactive-mode-map
        (copy-keymap comint-mode-map)))
(define-key jakarta-interactive-mode-map "|" 'jakarta-electric)
(define-key jakarta-interactive-mode-map ")" 'jakarta-electric-rp)
(define-key jakarta-interactive-mode-map "}" 'jakarta-electric-rc)
(define-key jakarta-interactive-mode-map "]" 'jakarta-electric-rb)
(define-key jakarta-interactive-mode-map "\t" 'jakarta-indent-command)
(define-key jakarta-interactive-mode-map "\C-c\C-i" 'jakarta-interrupt-jakarta)
(define-key jakarta-interactive-mode-map "\C-c\C-k" 'jakarta-kill-jakarta)
(define-key jakarta-interactive-mode-map "\C-c`"
  'jakarta-interactive-next-error-toplevel)
(define-key jakarta-interactive-mode-map "\C-m"
  'jakarta-interactive-send-input)
(define-key jakarta-interactive-mode-map "\C-j"
  'jakarta-interactive-send-input-or-indent)
(define-key jakarta-interactive-mode-map "\M-\C-m"
  'comint-send-input)
(if (functionp 'read-kbd-macro)
    (define-key jakarta-interactive-mode-map (read-kbd-macro "<kp-enter>")
      'comint-send-input))

(defconst jakarta-interactive-buffer-name "*jakarta-toplevel*")

(defconst jakarta-interactive-toplevel-error-regexp
  "[ \t]*Characters \\([0-9]+\\)-\\([0-9]+\\):"
  "Regular expression extracting the character numbers
from an error message produced by jakarta toplevel.")
(defvar jakarta-interactive-last-phrase-pos-in-source 0)
(defvar jakarta-interactive-last-phrase-pos-in-toplevel 0)

(defun jakarta-interactive-filter (text)
  (if (eq major-mode 'jakarta-interactive-mode)
      (save-excursion
	(if (< comint-last-input-end comint-last-input-start) ()
	  (if (and jakarta-with-xemacs jakarta-interactive-read-only-input)
	      (add-text-properties
	       comint-last-input-start comint-last-input-end
	       (list 'read-only t)))
	  (if (and (or jakarta-window-system jakarta-with-xemacs)
		   (featurep 'font-lock)
		   jakarta-interactive-input-font-lock)
	      (progn
		(font-lock-fontify-region comint-last-input-start
					  comint-last-input-end)
		(if (featurep 'sym-lock)
		    (sym-lock-make-symbols-atomic comint-last-input-start
						  comint-last-input-end))))
	  (if jakarta-interactive-output-font-lock
	      (save-excursion
		(goto-char (point-max))
		(re-search-backward comint-prompt-regexp
				    comint-last-input-end t)
		(add-text-properties
		 comint-last-input-end (point)
		 '(face jakarta-font-lock-interactive-output-face))))
	  (if jakarta-interactive-error-font-lock
	      (save-excursion
		(goto-char comint-last-input-end)
		(while (re-search-forward jakarta-interactive-error-regexp () t)
		  (let ((matchbeg (match-beginning 1))
			(matchend (match-end 1)))
		    (save-excursion
		      (add-text-properties
		       matchbeg matchend
		       '(face jakarta-font-lock-interactive-error-face))
		      (goto-char matchbeg)
		      (if (looking-at jakarta-interactive-toplevel-error-regexp)
			  (let ((beg (string-to-int (jakarta-match-string 1)))
				(end (string-to-int (jakarta-match-string 2))))
			    (add-text-properties
			     (+ comint-last-input-start beg)
			     (+ comint-last-input-start end)
			     '(face text-cursor)))))))))))))

(defun jakarta-interactive-mode ()
  "Major mode for interacting with a Jakarta process.
Runs a Jakarta toplevel as a subprocess of Emacs, with I/O through an
Emacs buffer. A history of input phrases is maintained. Phrases can
be sent from another buffer in Jakarta mode.

Special keys for Jakarta interactive mode:\\{jakarta-interactive-mode-map}"
  (interactive)
  (comint-mode)
  (if (not (eq (jakarta-install-font-lock t) 'font-lock)) ()
    (add-hook 'comint-output-filter-functions 'jakarta-interactive-filter)
    (if (not (boundp 'after-change-functions)) ()
      (make-local-hook 'after-change-functions)
      (put 'after-change-functions 'permanent-local t)
      (remove-hook 'after-change-functions 'font-lock-after-change-function t))
    (if (not (boundp 'pre-idle-hook)) ()
      (make-local-hook 'pre-idle-hook)
      (put 'pre-idle-hook 'permanent-local t)
      (remove-hook 'pre-idle-hook 'font-lock-pre-idle-hook t)))
  (setq comint-prompt-regexp "^#  *")
  (setq comint-process-echoes nil)
  (setq comint-get-old-input 'jakarta-interactive-get-old-input)
  (setq major-mode 'jakarta-interactive-mode)
  (setq mode-name "Jakarta-Interactive")
  (setq comint-scroll-to-bottom-on-output t)
  (use-local-map jakarta-interactive-mode-map)
  (set-syntax-table jakarta-mode-syntax-table)
  (setq local-abbrev-table jakarta-mode-abbrev-table)

  (if jakarta-window-system (jakarta-interactive-build-menu))

  ;; hooks for jakarta-interactive-mode
  (run-hooks 'jakarta-interactive-mode-hook))

(defun jakarta-run-jakarta (&optional cmd)
  "Run a Jakarta toplevel process. I/O via buffer `*jakarta-toplevel*'."
  (interactive
   (list (if (not (comint-check-proc jakarta-interactive-buffer-name))
	     (read-from-minibuffer "Jakarta toplevel to run: "
				   jakarta-interactive-program))))
  ;;(jakarta-run-process-if-needed cmd)
  (when jakarta-display-buffer-on-eval
    (display-buffer jakarta-interactive-buffer-name)))

(defun jakarta-run-process-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc jakarta-interactive-buffer-name)
	  (setq cmd jakarta-interactive-program)
	(setq cmd (read-from-minibuffer "Jakarta toplevel to run: "
					jakarta-interactive-program))))
  (setq jakarta-interactive-program cmd)
  (if (not (comint-check-proc jakarta-interactive-buffer-name))
      (let ((cmdlist (jakarta-args-to-list cmd))
            (process-connection-type nil))
	(set-buffer (apply (function make-comint) "jakarta-toplevel"
			   (car cmdlist) nil (cdr cmdlist)))
	(jakarta-interactive-mode)
	(sleep-for 1))))

(defun jakarta-jakartadoc-jakarta (&optional cmd)
  "Run a Jakarta toplevel process. I/O via buffer `*jakarta-toplevel*'."
  (interactive
   (list (if (not (comint-check-proc jakarta-interactive-buffer-name))
	     (read-from-minibuffer "JaKarTadoc run: "
				 jakarta-jakartadoc-program 
					
				   ))))
  (jakarta-jakartadoc-process-if-needed cmd)
  (when jakarta-display-buffer-on-eval
    (display-buffer jakarta-interactive-buffer-name)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun jakarta-process-buffer-name (name)
  "Return name of JaKarTa buffer associated with the document NAME."
  (concat "*" (abbreviate-file-name (expand-file-name name)) " output*"))

(defun jakarta-process-buffer (name)
  "Return the JaKarTa buffer associated with the document NAME."
  (get-buffer (jakarta-process-buffer-name name)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun jakarta-jakartadoc-process-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc jakarta-interactive-buffer-name)
	  (setq cmd jakarta-jakartadoc-program)
	(setq cmd (read-from-minibuffer "JaKarTadoc "
					jakarta-jakartadoc-program))))
  (setq jakarta-jakartadoc-program cmd)
  (if (not (comint-check-proc jakarta-interactive-buffer-name))
      (let ((cmdlist (jakarta-args-to-list cmd))
            (process-connection-type nil))
	(set-buffer (apply (function make-comint) "jakarta-toplevel"
			   (car cmdlist) nil (cdr cmdlist)))
	(jakarta-interactive-mode)
	(sleep-for 1))))

(defun jakarta-compile-jakarta (&optional cmd)
  "Run a Jakarta toplevel process. I/O via buffer `*jakarta-toplevel*'."
  (interactive
   (list (if (not (comint-check-proc jakarta-interactive-buffer-name))
	     (read-from-minibuffer "Jakarta toplevel to run: "
				   (concat jakarta-compile-program 
					   (buffer-file-name))
	   ))))
  (jakarta-compile-process-if-needed cmd)
  (when jakarta-display-buffer-on-eval
    (display-buffer jakarta-interactive-buffer-name)))

(defun jakarta-compile-process-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc jakarta-interactive-buffer-name)
	  (setq cmd jakarta-interactive-program)
	(setq cmd (read-from-minibuffer "Jakarta toplevel to run: "
					jakarta-compile-program))))
  (setq jakarta-compile-program cmd)
  (if (not (comint-check-proc jakarta-interactive-buffer-name))
      (let ((cmdlist (jakarta-args-to-list cmd))
            (process-connection-type nil))
	(set-buffer (apply (function make-comint) "jakarta-toplevel"
			   (car cmdlist) nil (cdr cmdlist)))
	(jakarta-interactive-mode)
	(sleep-for 1))))

(defun jakarta-args-to-list (string)
  (let ((where (string-match "[ \t]" string)))
    (cond ((null where) (list string))
	  ((not (= where 0))
	   (cons (substring string 0 where)
		 (jakarta-args-to-list (substring string (+ 1 where)
						 (length string)))))
	  (t (let ((pos (string-match "[^ \t]" string)))
	       (if (null pos)
		   nil
		 (jakarta-args-to-list (substring string pos
						 (length string)))))))))

(defun jakarta-interactive-get-old-input ()
  (save-excursion
    (let ((end (point)))
      (re-search-backward comint-prompt-regexp (point-min) t)
      (if (looking-at comint-prompt-regexp)
	  (re-search-forward comint-prompt-regexp))
      (buffer-substring-no-properties (point) end))))

(defun jakarta-interactive-end-of-phrase ()
  (save-excursion
    (end-of-line)
    (jakarta-find-meaningful-word)
    (jakarta-find-meaningful-word)
    (looking-at ".")))

(defconst jakarta-interactive-send-warning
  "Note: toplevel processing requires a terminating `.'")
(defun jakarta-interactive-send-input ()
  "Process if the current line ends with `.' then send the
current phrase else insert a newline."
  (interactive)
  (if (jakarta-interactive-end-of-phrase)
      (progn
	(comint-send-input)
	(goto-char (point-max)))
    (insert "\n")
    (message jakarta-interactive-send-warning)))

(defun jakarta-interactive-send-input-or-indent ()
  "Process if the current line ends with `.' then send the
current phrase else insert a newline and indent."
  (interactive)
  (if (jakarta-interactive-end-of-phrase)
      (progn
	(comint-send-input)
	(goto-char (point-max)))
    (insert "\n")
    (jakarta-indent-command)
    (message jakarta-interactive-send-warning)))

(defun jakarta-eval-region (start end)
  "Eval the current region in the Jakarta toplevel."
  (interactive "r")
  (save-excursion (jakarta-run-process-if-needed))
  (comint-preinput-scroll-to-bottom)
  (setq jakarta-interactive-last-phrase-pos-in-source start)
  (save-excursion
    (goto-char start)
    (jakarta-skip-blank-and-comments)
    (setq start (point))
    (goto-char end)
    (jakarta-skip-to-end-of-phrase)
    (setq end (point))
    (let ((text (buffer-substring-no-properties start end)))
      (goto-char end)
      (if (string= text "")
	  (message "Cannot send empty commands to Jakarta toplevel!")
	(set-buffer jakarta-interactive-buffer-name)
	(goto-char (point-max))
	(setq jakarta-interactive-last-phrase-pos-in-toplevel (point))
	(if jakarta-interactive-echo-phrase
	    (progn
	      (insert (concat text " "))
	      (comint-send-input))
	  (comint-send-string jakarta-interactive-buffer-name (concat text "."))
	  (comint-send-input))))
    (when jakarta-display-buffer-on-eval
      (display-buffer jakarta-interactive-buffer-name))))

(defun jakarta-eval-phrase ()
  "Eval the surrounding Jakarta phrase (or block) in the Jakarta toplevel."
  (interactive)
  (let ((end))
    (save-excursion
      (let ((pair (jakarta-discover-phrase)))
	(setq end (nth 2 pair))
	(jakarta-eval-region (nth 0 pair) (nth 1 pair))))
    (if jakarta-skip-after-eval-phrase
	(goto-char end))))

(defun jakarta-eval-buffer ()
  "Send the buffer to the Jakarta Interactive process."
  (interactive)
  (jakarta-eval-region (point-min) (point-max)))

(defun jakarta-interactive-next-error-source ()
  (interactive)
  (let ((error-pos) (beg 0) (end 0))
    (save-excursion
      (set-buffer jakarta-interactive-buffer-name)
      (goto-char jakarta-interactive-last-phrase-pos-in-toplevel)
      (setq error-pos
	    (re-search-forward jakarta-interactive-toplevel-error-regexp
			       (point-max) t))
      (if error-pos
	  (progn
	    (setq beg (string-to-int (jakarta-match-string 1))
		  end (string-to-int (jakarta-match-string 2))))))
    (if (not error-pos)
	(message "No syntax or typing error in last phrase.")
      (setq beg (+ jakarta-interactive-last-phrase-pos-in-source beg)
	    end (+ jakarta-interactive-last-phrase-pos-in-source end))
      (goto-char beg) (push-mark end t t))))

(defun jakarta-interactive-next-error-toplevel ()
  (interactive)
  (let ((error-pos) (beg 0) (end 0))
    (save-excursion
      (goto-char jakarta-interactive-last-phrase-pos-in-toplevel)
      (setq error-pos
	    (re-search-forward jakarta-interactive-toplevel-error-regexp
			       (point-max) t))
      (if error-pos
	  (setq beg (string-to-int (jakarta-match-string 1))
		end (string-to-int (jakarta-match-string 2)))))
    (if (not error-pos)
	(message "No syntax or typing error in last phrase.")
      (setq beg (+ jakarta-interactive-last-phrase-pos-in-toplevel beg)
	    end (+ jakarta-interactive-last-phrase-pos-in-toplevel end))
      (goto-char beg) (push-mark end t t))))

(defun jakarta-interrupt-jakarta ()
  (interactive)
  (if (comint-check-proc jakarta-interactive-buffer-name)
      (save-excursion
	(set-buffer jakarta-interactive-buffer-name)
	(comint-interrupt-subjob))))

(defun jakarta-kill-jakarta ()
  (interactive)
  (if (comint-check-proc jakarta-interactive-buffer-name)
      (save-excursion
	(set-buffer jakarta-interactive-buffer-name)
	(comint-kill-subjob))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Menu support

(defun jakarta-about () (interactive)
  (describe-variable 'jakarta-mode-version))
(defun jakarta-help () (interactive)
  (describe-function 'jakarta-mode))
(defun jakarta-interactive-help () (interactive)
  (describe-function 'jakarta-interactive-mode))

(defvar jakarta-definitions-menu-last-buffer nil)
(defvar jakarta-definitions-keymaps nil)

(defun jakarta-build-menu ()
  (if (condition-case nil (prog2 (require 'easymenu) nil) (error t)) ()
    (easy-menu-define
     jakarta-mode-menu (list jakarta-mode-map)
     "Xanana Mode Menu."
     '("Xanana"
       ("Execution Mode"
	["Run Jakarta compile" jakarta-compile-jakarta t]
	["Run Jakarta Toplevel" jakarta-run-jakarta t]
	["Interrupt Jakarta Toplevel" jakarta-interrupt-jakarta
	 :active (comint-check-proc jakarta-interactive-buffer-name)]
	["Kill Jakarta Toplevel" jakarta-kill-jakarta
	 :active (comint-check-proc jakarta-interactive-buffer-name)]
	["Evaluate Region" jakarta-eval-region
	 ;; region-active-p for XEmacs and mark-active for Emacs
	 :active (if (fboundp 'region-active-p) (region-active-p) mark-active)]
	["Evaluate Phrase" jakarta-eval-phrase t]
	["Evaluate Buffer" jakarta-eval-buffer t])
       ("Jakarta Forms"
	["data.." jakarta-insert-data-form t]
	["definition.." jakarta-insert-definition-form t]
	["function.." jakarta-insert-function-form t]
	["record.." jakarta-insert-record-form t]
	["type.." jakarta-insert-type-form t]
	["variable.." jakarta-insert-variable-form t]
	 
	
	 "---"
         ["abstract..with" jakarta-insert-abstract-with-form t]
	 ["build..from.." jakarta-insert-build-from-form t]
	 "---"
	 ["coercion.." jakarta-insert-coercion-form t]
	 ["delete.." jakarta-insert-delete-form t]
	 ["delete..in.." jakarta-insert-delete-in-form t]
	 ["determine.." jakarta-insert-determine-form t]
	 ["drop.." jakarta-insert-drop-form t]
	 ["into..log.." jakarta-insert-into-log-form t]
	 ["reduce..at.." jakarta-insert-reduce-form t]
	 ["reject..in.." jakarta-insert-reject-in-form t]
	 ["remove..in.." jakarta-insert-remove-in-form t]
	 ["replace..by.." jakarta-insert-replace-by-form t]
	 ["replace..by..in.." jakarta-insert-replace-by-in-form t]
	 ["replace_type.." jakarta-insert-replace_type-form t]
	 ["select.." jakarta-insert-select-form t]
	 ["select..in" jakarta-insert-select-in-form t]
	 ["totalize.." jakarta-insert-totalize-form t]
	 )
       ["Switch .dj/.dj" jakarta-find-alternate-file t]
       "---"
       ["Compile..." compile t]
       ["Reference Manual..." jakarta-browse-manual t]
       ["Jakarta Library..." jakarta-browse-library t]
       ("Definitions"
	["Scan..." jakarta-list-definitions t])
       "---"
       ["Customize Jakarta Mode..." (customize-group 'jakarta) t]
       ("Jakarta Options" ["Dummy" nil t])
       ("Jakarta Interactive Options" ["Dummy" nil t])
       "---"
       ["JaKarTadoc" jakarta-jakartadoc-jakarta t]
       "---"
       ["About" jakarta-about t]
       ["Help" jakarta-help t]))
    (easy-menu-add jakarta-mode-menu)
    (jakarta-update-options-menu)
    ;; save and update definitions menu
    (make-local-variable 'jakarta-definitions-menu)
    (if jakarta-with-xemacs
	(add-hook 'activate-menubar-hook 'jakarta-update-definitions-menu)
      (if (not (functionp 'easy-menu-create-keymaps)) ()
	;; patch for Emacs
	(add-hook 'menu-bar-update-hook
		  'jakarta-with-emacs-update-definitions-menu)
	(make-local-variable 'jakarta-definitions-keymaps)
	(setq jakarta-definitions-keymaps
	      (cdr (easy-menu-create-keymaps
		    "Definitions" jakarta-definitions-menu)))
	(setq jakarta-definitions-menu-last-buffer nil)))))

(defun jakarta-interactive-build-menu ()
  (if (condition-case nil (prog2 (require 'easymenu) nil) (error t))
      ()
    (easy-menu-define
     jakarta-interactive-mode-menu (list jakarta-interactive-mode-map)
     "Xanana Interactive Mode Menu."
     '("Xanana"
       ("Execution Mode"
	["Run Jakarta compile" jakarta-compile-jakarta t]
	["Run Jakarta Toplevel" jakarta-run-jakarta t]
	["Interrupt Jakarta Toplevel" jakarta-interrupt-jakarta
	 :active (comint-check-proc jakarta-interactive-buffer-name)]
	["Kill Jakarta Toplevel" jakarta-kill-jakarta
	 :active (comint-check-proc jakarta-interactive-buffer-name)]
	["Evaluate Region" jakarta-eval-region :active (region-active-p)]
	["Evaluate Phrase" jakarta-eval-phrase t]
	["Evaluate Buffer" jakarta-eval-buffer t])
       "---"
       ["Customize Jakarta Mode..." (customize-group 'jakarta) t]
       ("Jakarta Options" ["Dummy" nil t])
       ("Jakarta Interactive Options" ["Dummy" nil t])
       "---"
       ["JaKarTadoc" jakarta-jakartadoc-jakarta t]
       "---"
       ["About" jakarta-about t]
       ["Help" jakarta-interactive-help t]))
    (easy-menu-add jakarta-interactive-mode-menu)
    (jakarta-update-options-menu)))

(defun jakarta-update-definitions-menu ()
  (if (eq major-mode 'jakarta-mode)
      (easy-menu-change
       '("Jakarta") "Definitions"
       jakarta-definitions-menu)))

(defun jakarta-with-emacs-update-definitions-menu ()
  (if (current-local-map)
      (let ((keymap
	     (lookup-key (current-local-map) [menu-bar Jakarta Definitions])))
	(if (and
	     (keymapp keymap)
	     (not (eq jakarta-definitions-menu-last-buffer (current-buffer))))
	    (setcdr keymap jakarta-definitions-keymaps)
	  (setq jakarta-definitions-menu-last-buffer (current-buffer))))))

(defun jakarta-toggle-option (symbol)
  (interactive)
  (set symbol (not (symbol-value symbol)))
  (if (eq 'jakarta-use-abbrev-mode symbol)
      (abbrev-mode jakarta-use-abbrev-mode)) ; toggle abbrev minor mode
  (if jakarta-with-xemacs nil (jakarta-update-options-menu)))

(defun jakarta-update-options-menu ()
  (easy-menu-change
   '("Jakarta") "Jakarta Options"
   (mapcar (lambda (pair)
	     (if (consp pair)
		 (vector (car pair)
			 (list 'jakarta-toggle-option (cdr pair))
			 ':style 'toggle
			 ':selected (nth 1 (cdr pair))
			 ':active t)
	       pair)) jakarta-options-list))
  (easy-menu-change
   '("Jakarta") "Jakarta Interactive Options"
   (mapcar (lambda (pair)
	     (if (consp pair)
		 (vector (car pair)
			 (list 'jakarta-toggle-option (cdr pair))
			 ':style 'toggle
			 ':selected (nth 1 (cdr pair))
			 ':active t)
	       pair)) jakarta-interactive-options-list)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Browse Manual

;; From M. Quercia

(defun jakarta-browse-manual ()
  "*Browse Jakarta reference manual."
  (interactive)
  (setq jakarta-manual-url (read-from-minibuffer "URL: " jakarta-manual-url))
  (funcall jakarta-browser jakarta-manual-url))

(defun jakarta-xemacs-w3-manual (url)
  "*Browse Jakarta reference manual."
  (w3-fetch-other-frame url))

(defun jakarta-netscape-manual (url)
  "*Browse Jakarta reference manual."
  (start-process-shell-command
   "netscape" nil
   (concat "netscape -remote 'openURL ("
	   url ", newwindow)' || netscape " url)))

(defun jakarta-mmm-manual (url)
  "*Browse Jakarta reference manual."
  (start-process-shell-command
   "mmm" nil
   (concat "mmm_remote " url " || mmm -external " url)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Browse Library

;; From M. Quercia

(defun jakarta-browse-library()
  "Browse the Jakarta library."
  (interactive)
  (let ((buf-name "*jakarta-library*") (opoint)
	(dir (read-from-minibuffer "Library path: " jakarta-library-path)))
    (if (and (file-directory-p dir) (file-readable-p dir))
	(progn
	  (setq jakarta-library-path dir)
	  ;; list *.ml and *.dj files
	  (with-output-to-temp-buffer buf-name
	    (buffer-disable-undo standard-output)
	    (save-excursion
	      (set-buffer buf-name)
	      (kill-all-local-variables)
	      (make-local-variable 'jakarta-library-path)
	      (setq jakarta-library-path dir)
	      ;; help
	      (insert "Directory \"" dir "\".\n") 
	      (insert "Select a file with middle mouse button or RETURN.\n\n")
	      (insert "Interface files (.dj):\n\n")
	      (insert-directory (concat dir "/*.dj") "-C" t nil)
	      (insert "\n\nImplementation files (.dj):\n\n")
	      (insert-directory (concat dir "/*.dj") "-C" t nil)
	      ;; '.', '-' and '_' are now letters
	      (modify-syntax-entry ?. "w")
	      (modify-syntax-entry ?_ "w")
	      (modify-syntax-entry ?- "w")
	      ;; every file name is now mouse-sensitive
	      (goto-char (point-min))
	      (while (< (point) (point-max))
		(re-search-forward "\\.dj.?\\>")
		(setq opoint (point))
		(re-search-backward "\\<" (point-min) 1)
		(put-text-property (point) opoint 'mouse-face 'highlight)
		(goto-char (+ 1 opoint)))
	      ;; activate jakarta-library mode
	      (setq major-mode 'jakarta-library-mode)
	      (setq mode-name "jakarta-library")
	      (use-local-map jakarta-library-mode-map)
	      (setq buffer-read-only t)))))))
  
(setq jakarta-library-mode-map (make-keymap))
(suppress-keymap jakarta-library-mode-map)
(define-key jakarta-library-mode-map [return] 'jakarta-library-find-file)
(define-key jakarta-library-mode-map [mouse-2] 'jakarta-library-mouse-find-file)
  
(defun jakarta-library-find-file ()
  "Load the file whose name is near point."
  (interactive)
  (save-excursion
    (if (text-properties-at (point))
	(let (beg)
	  (re-search-backward "\\<") (setq beg (point))
	  (re-search-forward "\\>")
	  (find-file-read-only (concat jakarta-library-path "/"
				       (buffer-substring-no-properties
					beg (point))))))))

(defun jakarta-library-mouse-find-file (event)
  "Visit the file name you click on."
  (interactive "e")
  (let ((owindow (selected-window)))
    (mouse-set-point event)
    (jakarta-library-find-file)
    (select-window owindow)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Definitions List

;; Designed from original code by M. Quercia

;; (defconst jakarta-definitions-regexp
;;  "\\<\\(and\\|val\\|value\\|type\\|module\\|class\\|exception\\|let\\)\\>"
;;  "Regexp matching definition phrases.")

(defconst jakarta-definitions-regexp
  "\\<\\(data\\|definition\\|type\\|declaration\\|module\\|variable\\|record\\|function\\)\\>"
  "Regexp matching definition phrases.")

(defconst jakarta-definitions-bind-skip-regexp
  "\\<\\(\\)\\>\\|'[A-Za-z\300-\377][0-9_'A-Za-z\300-\377]*\\|('.*)"
  "Regexp matching stuff to ignore after a binding keyword.")

(defvar jakarta-definitions-menu (list ["Scan..." jakarta-list-definitions t])
  "Initial content of the definitions menu.")
(make-variable-buffer-local 'jakarta-definitions-menu)

(defun jakarta-list-definitions ()
  "Parses the buffer and gathers toplevel definitions for quick
jump via the definitions menu'."
  (interactive)
  (message "Searching definitions...")
  (save-excursion
    (let ((cpt 0) (kw) (menu)
	  (value-list) (type-list) (module-list)  (misc-list))
      (goto-char (point-min))
      (jakarta-skip-blank-and-comments)
      (while (< (point) (point-max))
	(if (looking-at jakarta-definitions-regexp)
	    (progn
	      (setq kw (jakarta-match-string 0))
	     ;; (if (string= kw "and")
		;;  (setq kw (save-match-data
			;;     (save-excursion (jakarta-find-and-match)))))
	      (if (or (string= kw "data")
		     ;; (string= kw "val")
		      (string= kw "record")) (setq kw "type"))
	       (if (or (string= kw "definition")
		       (string= kw "variable")
		      (string= kw "declaration")) (setq kw "function"))
	      ;; skip optional elements
	      (goto-char (match-end 0))
	      (jakarta-skip-blank-and-comments)
	      (if (looking-at jakarta-definitions-bind-skip-regexp)
		  (goto-char (match-end 0)))
	      (jakarta-skip-blank-and-comments)
	      (if (looking-at "\\<[A-Za-z\300-\377][0-9_'A-Za-z\300-\377]*\\>")
		  ;; menu item : [name (goto-char ...) t]
		  (let* ((p (make-marker))
			 (ref (vector (jakarta-match-string 0)
				      (list 'jakarta-goto p) t)))
		    (setq cpt (1+ cpt))
		    (message (concat "Searching definitions... ("
				     (number-to-string cpt) ")"))
		    (set-marker p (point))
		    (cond
		     ((string= kw "function")
		      (setq value-list (cons ref value-list)))
		     ((string= kw "type")
		      (setq type-list (cons ref type-list)))
		     ((string= kw "module")
		      (setq module-list (cons ref module-list)))
		     ;;((string= kw "class")
		     ;; (setq class-list (cons ref class-list)))
		     (t (setq misc-list (cons ref misc-list))))))))
	;; skip to next phrase or next top-level `and'
	(jakarta-forward-char)
;;	(jakarta-next-phrase t)
	(let ((old-point (point)) (last-and))
	  (setq last-and (point))
	 ;; (save-excursion
	   ;; (while (and (re-search-backward "\\<and\\>" old-point t)
		;;	(not (jakarta-in-literal-or-comment-p))
		;;	(save-excursion (jakarta-find-and-match)
		;;			(>= old-point (point))))
	     ;; (setq last-and (point))))
	  (goto-char last-and)))
      ;; sort and build lists
      (mapcar (lambda (pair)
		(if (cdr pair)
		    (setq menu
			  (append (jakarta-split-long-list
			    (car pair) (jakarta-sort-definitions (cdr pair)))
				  menu))))
	      (list (cons "Miscellaneous" misc-list)
		    (cons "Values" value-list)
		   ;; (cons "Classes" class-list)
		    (cons "Types" type-list)
		    (cons "Modules" module-list)))
;;classes tirar ver;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;,
      ;; update definitions menu
      (setq jakarta-definitions-menu
	    (append menu (list "---" ["Rescan..." jakarta-list-definitions t])))
      (if (or jakarta-with-xemacs
	      (not (functionp 'easy-menu-create-keymaps))) ()
	;; patch for Emacs 20.2
	(setq jakarta-definitions-keymaps
	      (cdr (easy-menu-create-keymaps 
		    "Definitions" jakarta-definitions-menu)))
	(setq jakarta-definitions-menu-last-buffer nil))
      (message "Searching definitions... done")))
  (jakarta-update-definitions-menu))

(defun jakarta-goto (pos)
  (goto-char pos)
  (recenter))

(defun jakarta-sort-definitions (list)
  (let* ((last "") (cpt 1)
	 (list (sort (nreverse list)
		     (lambda (p q) (string< (elt p 0) (elt q 0)))))
	 (tail list))
    (while tail
      (if (string= (elt (car tail) 0) last)
	  (prog2
	      (setq cpt (1+ cpt))
	      (aset (car tail) 0 (format "%s (%d)" last cpt)))
	(setq cpt 1)
	(setq last (elt (car tail) 0)))
      (setq tail (cdr tail)))
    list))

;; look for the (n-1)th or last element of a list
(defun jakarta-nth (n list)
  (if (or (<= n 1) (null list) (null (cdr list))) list
    (jakarta-nth (1- n) (cdr list))))
    
;; split a definition list if it is too long
(defun jakarta-split-long-list (title list)
  (let ((tail (jakarta-nth jakarta-definitions-max-items list)))
    (if (or (null tail) (null (cdr tail)))
        ;; list not too long, cons the title
        (list (cons title list))
      ;; list too long, split and add initials to the title
      (let (lists)
        (while list
          (let ((beg (substring (elt (car list) 0) 0 1))
                (end (substring (elt (car tail) 0) 0 1)))
            (setq lists (cons
                         (cons (format "%s %s-%s" title beg end) list)
                         lists))
            (setq list (cdr tail))
            (setcdr tail nil)
            (setq tail (jakarta-nth jakarta-definitions-max-items list))))
        (nreverse lists)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Hooks and Exit

(defvar jakarta-load-hook nil
  "This hook is run when Jakarta is loaded in. It is a good place to put
key-bindings or hack Font-Lock keywords...")

(run-hooks 'jakarta-load-hook)

(provide 'caml) ;; for compatibility with jakarta support modes
                ;; you may also link jakarta.el to jakarta.el
(provide 'jakarta)
